#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.base import Base
from appy.model.fields import Show
from appy.model.query import Query
from appy.xml.escape import Escape
from appy.model.fields.pod import Pod
from appy.model.fields.rich import Rich
from appy.model.fields.file import File
from appy.model.fields.info import Info
from appy.model.document import Document
from appy.model.carousel import Carousel
from appy.model.fields.group import Group
from appy.ui.layout import Layout, Layouts
from appy.model.fields.select import Select
from appy.model.fields.string import String
from appy.model.fields.boolean import Boolean
from appy.model.fields.integer import Integer
from appy.model.fields.ref import Ref, autoref
from appy.model.fields.phase import Page as FPage
from appy.model.workflow.standard import Anonymous

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
EXPR_ERR = 'Page "%s" (%s): error while evaluating page expression "%s" (%s).'
DELETED  = 'Web page %s deleted.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Page(Base):
    '''Base class representing a web page'''

    # By default, web pages are public
    workflow = Anonymous

    pa = {'label': 'Page'}
    # Pages are not indexed by default
    indexable = False

    @staticmethod
    def getListColumns(tool):
        '''Show minimal info for the anonymous user'''
        if tool.user.isAnon(): return
        return ('title', 'expression', 'next', 'selectable*100px|',
                'podable*80px|', 'viewCss*100px|')
    listColumns = getListColumns

    # The POD ouput
    doc = Pod(template='/model/pod/Page.odt', formats=('pdf',), show=False,
              layouts=Pod.Layouts.inline, freezeTemplate=lambda o,tpl: ('pdf',))

    @staticmethod
    def update(class_):
        '''Configure field "title"'''
        title = class_.fields['title']
        title.show = Show.EX
        title.label = 'Page'
        title.page.show = lambda o: True if o.allows('write') else 'view'
        title.page.sticky = True

    # The POD output appears inline in the sub-breadcrumb
    def getSubBreadCrumb(self):
        '''Display an icon for downloading this page (and sub-pages if any) as a
           POD.'''
        if self.podable: return self.getField('doc').doRender('view', self)

    def getSubTitle(self):
        '''Render a link to the tied carousel and/or search'''
        r = []
        esc = Escape.xhtml
        for name in ('carousel', 'query'):
            if not self.isEmpty(name):
                o = getattr(self, name)
                label = self.translate(self.getField(name).labelId)
                r.append('<div class="discreet">%s: <a href="%s">%s</a></div>' \
                         % (esc(label), o.url, esc(o.title)))
        return '\n'.join(r)

    # A warning: image upload is impossible while the page is temp
    warning = Info(show=lambda o: 'edit' if o.isTemp() else None,
                   focus=True, layouts=Info.Layouts.n, **pa)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            Linked carousel
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def showCarousel(self):
        '''Show field "carousel" if carousels are defined'''
        if self.tool.isEmpty('carousels'): return
        # Do not show it (except on "edit") if there is no tied carousel
        return True if not self.isEmpty('carousel') else 'edit'

    carousel = Ref(Carousel, add=False, link=True, render='links',
      select=lambda o: o.tool.carousels, shownInfo=Carousel.pageListColumns,
      show=showCarousel, layouts=Layouts.fvd,
      view=Px('''<x var="o=o.carousel">:o.pxView</x>'''),
      back=Ref(attribute='pages', multiplicity=(0,None), render='links',
               group=Carousel.mainGroup, label='Carousel'), **pa)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            Page content
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    content = Rich(documents='documents', height='350px', inject=True,
                   viewCss=lambda o: o.viewCss, layouts=Rich.Layouts.f, **pa)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                           Linked search
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def showQuery(self):
        '''Show field "query" if queries are defined'''
        if self.tool.isEmpty('queries'): return
        # Do not show it (except on "edit") if there is no tied carousel
        return True if not self.isEmpty('query') else 'edit'

    query = Ref(Query, add=False, link=True, render='links',
      select=lambda o: o.tool.queries, shownInfo=Query.pageListColumns,
      show=showQuery, layouts=Layouts.fvd,
      view=Px('''<x var="o=o.query">:o.pxView</x>'''),
      back=Ref(attribute='pages', multiplicity=(0,None), render='links',
               group=Query.mainGroup, label='Query'), **pa)

    def forRootOnEdit(self):
        '''Determines visibility of some fields, on the "edit" layout, for root
           pages only.'''
        return Show.V_ if self.isRoot() else None

    # Is this (root) page selectable in the main dropdown ?
    selectable = Boolean(default=True, show=forRootOnEdit,
                         layouts=Boolean.Layouts.d, **pa)

    # Is PDF POD export enabled for this page ? Sub-pages can't be exported.
    podable = Boolean(layouts=Boolean.Layouts.d, show=forRootOnEdit, **pa)

    # Maximum width for images uploaded and rendered within the XHTML content
    maxWidth = Integer(default=700, show=False)

    # If this Python expression returns False, the page can't be viewed
    def showExpression(self):
        '''Show the expression to managers only'''
        # Do not show it on "view" if empty
        if self.isEmpty('expression'): return Show.V_
        return self.allows('write')

    expression = String(layouts=Layouts.d, show=showExpression, **pa)

    # Name of the CSS class applied to the page content, when rendered on view
    # or cell layouts.
    viewCss = String(default='xhtml', layouts=Layouts.d, show=Show.V_, **pa)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                           Inner images
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # The images (or other documents) that may be included in field "content"
    documents = Ref(Document, add=True, link=False, multiplicity=(0,None),
      composite=True, back=Ref(attribute='page', show=False, label='Document'),
      showHeaders=True, shownInfo=Document.listColumns, actionsDisplay='inline',
      page=FPage('images', show=lambda o:'view' if o.allows('write') else None),
      rowAlign='middle', **pa)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                         Background image
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # In most cases, when a page is shown, like any other object, the background
    # image possibly applied is baseBG.jpg. On public root pages, no background
    # is shown: the one uploaded here may be used instead.
    bip = {'page': FPage('bg', show=lambda o: o.allows('write') and o.isRoot()),
           'group': Group('main', style='grid', hasLabel=False),
           'layouts': Layouts.gd }
    bip.update(pa)

    backgroundImage = File(isImage=True, viewWidth='700px', cache=True, **bip)

    # When the background image hereabove is used, the following parameter
    # determines how it is rendered. It is the equivalent of CSS attribute
    # "background-size".
    backgroundSize = Select(validator=('cover', 'auto'), default='cover', **bip)

    def initialiseLocalPage(self, req):
        '''A page can be specifically rendered as a root public page or as
           an inner page of such a root page. In these cases, the rendering
           context must be adapted.'''
        # p_self may propose a specific background image
        if not self.isEmpty('backgroundImage'):
            style = 'background-image:url(%s/backgroundImage/download); ' \
                    'background-repeat:no-repeat; background-size:%s' % \
                    (self.url, self.backgroundSize)
        else:
            style = None
        # Use a minimal page layout that only includes the "w" part (=widgets)
        req.pageLayout = Layout('w', style=style)
        # Avoid including global page elements: they have already been included
        # by the enclosing element.
        req.notGlobal = True

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            Next page
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Pages can be chained, in order to produce a single, composite page.
    # Chained pages must all be in the same container object.

    def listSiblings(self):
        '''Returns the pages being p_self's siblings'''
        return [page for page in self.container.pages if page != self]

    def hasContainer(self):
        '''Some fields can't be shown while a page has no container, which can
           happen when creating a page from the porlet.'''
        return Show.V_ if self.container else None

    next = Ref(None, add=False, link=True, layouts=Layouts.fvd, render='links',
               back=Ref(attribute='previous', render='links',
                        label='Page', show=Show.VE_),
               show=hasContainer, select=listSiblings, **pa)

    def getChain(self):
        '''Returns the list of pages being chained to p_self via "next" fields,
           recursively.'''
        r = [self]
        next = self.next
        return r + next.getChain() if next else r

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            Sub-pages
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # A page can contain sub-pages
    def showSubPages(self):
        '''For non-writers, show sub-pages only if present'''
        if self.allows('write'):
            # Show it, unless the page is rendered as the next page of a
            # previous one.
            return not self.req.notGlobal
        if not self.isEmpty('pages'): return 'view'

    pages = Ref(None, multiplicity=(0,None), add=True, link=False,
      composite=True, back=Ref(attribute='parent', show=False, **pa),
      showHeaders=True, actionsDisplay='inline', show=showSubPages,
      numbered=True, checkboxes=lambda o: o.allows('write'), **pa)

    def getParents(self, includeMyself=True):
        '''Returns, as a set, p_self's parents, including p_self itself if
           p_includeMyself is True.'''
        r = {self} if includeMyself else set()
        return r if self.isRoot() else r.union(self.container.getParents())

    def mayView(self):
        '''In addition to the workflow, evaluating p_self.expression, if
           defined, determines p_self's visibility.'''
        expression = self.expression
        if not expression: return True
        user = self.user
        try:
            return eval(expression)
        except Exception as err:
            message = EXPR_ERR % (self.title, self.id, expression, str(err))
            self.log(message, type='error')
            return True

    def getMergedContent(self, r=None, level=1):
        '''Returns a list of chunks of XHTML code, one for p_self and one for
           each of its sub-pages, recursively.'''
        # More precisely, the r_result is a list of tuples of the form
        #
        #                   ~(xhtmlChunk, i_delta)~
        #
        #, "delta" being the delta to apply to the title's outline levels,
        # depending on the depth of the page within the complete tree of pages.
        #
        # p_level is the current level of recursion, while p_r is the current
        # list under creation. Indeed, in order to avoid concatenating lists, at
        # each m_getMergedContent call, the current p_r(esult) is passed.
        if r is None:
            # We are at the start of a recursive merge: there is no result yet.
            # Create it.
            r = []
        # Two pieces of information must be collected about p_self: its title
        # and content.
        if level > 1:
            # At level 1, do not dump the title
            title = self.getValue('title', type='shown')
            r.append(('<h1>%s</h1>' % Escape.xhtml(title), level-2))
        # Add p_self's content
        if not self.isEmpty('content'):
            r.append((self.getValue('content', type='shown'), level-1))
        # Add sub-pages
        if not self.isEmpty('pages'):
            for page in self.pages:
                page.getMergedContent(r, level=level+1)
        return r

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Main methods
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def inPortlet(self, selected, level=0):
        '''Returns a chunk of XHTML code for representing this page in the
           portlet.'''
        # Is this page currently selected ? A selected page is the page being
        # currently consulted on the site or any of its parent pages.
        isSelected = self in selected
        # A specific CSS class wil be applied for a selected page
        css = ' class="current"' if isSelected else ''
        # If p_self has parents, we must render it with a margin-right
        style = ' style="margin-left:%dpx"' % (level * 5) if level else ''
        r = ['<div%s><a href="%s"%s>%s</a></div>' % \
             (style, self.url, css, Escape.xhtml(self.title))]
        if isSelected and not self.isEmpty('pages'):
            for sub in self.pages:
                r.append(sub.inPortlet(selected, level+1))
        return ''.join(r)

    def isRoot(self):
        '''Is p_self a root page ?'''
        # A page created from the portlet (if class Page is declared as root)
        # has no container when under creation.
        container = self.container
        return not container or (container.class_.name == 'Tool')

    def onEdit(self, created):
        '''Link the page among root pages if created from the portlet'''
        if created and not self.initiator:
            self.tool.link('pages', self)

    def onDelete(self):
        '''Log the page deletion'''
        self.log(DELETED % self.id)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  PXs
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # This selector allows to choose one root page among published tool.pages
    pxSelector = Px('''
      <select onchange="gotoURL(this)">
       <option value="">:_('goto_link')</option>
       <option for="page in pages"
               value=":'%s/public?rp=%s' % (tool.url, page.iid)"
               selected=":req.rp == str(page.iid)">:page.title</option>
      </select>''',

     js='''
       function gotoURL(select) {
         var url = select.value;
         if (url) goto(url);
       }''')

    # PX showing all root pages in the portlet, when shown for pages
    portletBottom = Px('''
     <div class="topSpaceS" var="pages=tool.OPage.getRoot(tool)">
      <x if="pages"
         var2="selected=o.getParents() if o.class_.name == 'Page' else {}">
       <x for="page in pages">::page.inPortlet(selected)</x>
      </x>
      <i if="not pages">:_('no_page')</i>
     </div>''')

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Class methods
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @classmethod
    def getRoot(class_, tool):
        '''Return the pages being visible by the logged user, among the site's
           root pages from p_tool.pages.'''
        # Return the cached version, if available
        cache = tool.H().cache
        if 'appyRootPages' in cache: return cache.appyRootPages
        # Compute it
        r = [page for page in tool.pages \
             if page.selectable and tool.guard.mayView(page)]
        cache.appyRootPages = r
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
autoref(Page, Page.next)
autoref(Page, Page.pages)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import datetime
from base64 import decodebytes

from DateTime import DateTime

from appy.xml import Parser
from appy.model.utils import Object as O

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
CONVERSION_ERROR = '"%s" value "%s" could not be converted by the XML ' \
                   'unmarshaller.'
CUSTOM_CONVERSION_ERROR = 'Custom converter for "%s" values produced an ' \
                          'error while converting value "%s". %s'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UnmarshalledFile:
    '''Used for reading XML-marshalled files and converting them to Python
       objects.'''

    def __init__(self):
        self.name = ''       # The name of the file on disk
        self.type = None     # The MIME type of the file
        self.value = b''     # The binary content of the file or a file object
        self.size = 0        # The length of the file in bytes

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Unmarshaller(Parser):
    '''Allows to parse a XML file and recreate the corresponding web of Python
       objects.'''

    # This parser assumes that the XML file respects this convention: any tag
    # may define in attribute "type" storing the type of its content, which may
    # be:

    #                   bool * int * float * long * bytes
    #              datetime * DateTime * tuple * list * object

    # If "object" is specified, it means that the tag contains sub-tags, each
    # one corresponding to the value of an attribute for this object. if "tuple"
    # is specified, it will be converted to a list. "long" is supported for
    # backward compatibility with Python 2.

    def __init__(self, classes={}, tagTypes={}, conversionFunctions={},
                 contentAttributes={}):
        Parser.__init__(self)
        # p_classes is a dict whose keys are tag names and values are Python
        # classes. During the unmarshalling process, when an object is
        # encountered within a given tag, if it corresponds to a key in this
        # dict, the marshaller will create a instance of this class. Else, an
        # instance of generic class appy.model.utils.Object will created
        # instead. Note that the constructor of classes specified in p_classes
        # will not be called: the marshaller will dynamically set the class to
        # the created instance.
        if not isinstance(classes, dict) and classes:
            # The user may only need to define a class for the root tag. The
            # default name for a root tag is "appyData". This is a naming
            # convention taken by the marshaller corresponding by this
            # unmarshaller, defined in appy/xml/marshaller.py.
            self.classes = {'appyData': classes}
        else:
            self.classes = classes

        # If the XML file to parse does not respect the conventions described
        # above, you can provide type information in p_tagTypes. Here is an
        # example of p_tagTypes:
        #
        #      {"information": "list", "days": "list", "person": "object"}
        #
        self.tagTypes = tagTypes

        # The parser assumes that data is represented in some standard way. If
        # it is not the case, you may provide, in this dict, custom functions
        # allowing to convert values of basic types (long, float, datetime...).
        # Every such function must take a single arg which is the value to
        # convert and return the converted value. Dict keys are strings
        # representing types ('bool', 'int', 'datetime', etc) and dict values
        # are conversion functions. Here is an example:
        #
        #        {'int': convertInteger, 'DateTime': convertDate}
        #
        # NOTE: you can even invent a new basic type, put it in self.tagTypes,
        # and create a specific conversionFunction for it. This way, you can
        # for example convert strings that have specific values (in this case,
        # knowing that the value is a "string" is not sufficient).
        self.conversionFunctions = conversionFunctions

        # If you have the special case of a tag whose type is "object" and that
        # has direct text content, you can specify, in attribute
        # "contentAttributes", the name of the attribute onto which this content
        # will be stored. For example, suppose you have:
        #
        #        <parent><child attr1="a1" attr2="a2">c1</child></parent>
        #
        # If you specify:
        #
        #                   tagTypes={'child': 'object'}
        #              contentAttributes={'child': 'content'}
        #
        # you will get, for "child", an object
        #
        #             O(attr1='a1', attr2='a2', content='c1')
        self.contentAttributes = contentAttributes

    def convertAttrs(self, attrs):
        '''Converts XML attrs to a dict'''
        r = {}
        for k, v in attrs.items():
            if ':' in k: # An attr prefixed with a namespace. Remove this.
                k = k.split(':')[-1]
            r[str(k)] = v
        return r

    def getDatetime(self, value):
        '''From this string p_value representing a date, produce an instance of
           datetime.datime.'''
        r = value.split()
        # First part should be the date part
        year, month, day = r[0].split('/')
        return datetime.datetime(int(year), int(month), int(day))

    def startDocument(self):
        # Standard parser attribute "r" will contain the resulting web of
        # unmarshalled Python objects.
        Parser.startDocument(self)
        # This attribute will store the name of the root tag
        self.rootTag = None
        # The stack of current "containers" where to store the next parsed
        # element. A container can be a list, a tuple or an object (the root
        # object of the whole web or a sub-object).
        self.env.containerStack = []
        # This attribute holds the name of the currently parsed basic type
        # (string, float...)
        self.env.currentBasicType = None
        # This attribute stores the content of the currently walked tag
        self.env.currentContent = ''

    containerTags = ('tuple', 'list', 'dict', 'object', 'file')
    numericTypes = ('bool', 'int', 'long', 'float')

    def startElement(self, tag, attrs):
        # Remember the name of the previous element
        previousTag = None
        if self.env.currentTag:
            previousTag = self.env.currentTag.name
        else:
            # We are walking the root tag
            self.rootTag = tag
        e = Parser.startElement(self, tag, attrs)
        # Determine the type of the element (the default is "str")
        type = 'str'
        if 'type' in attrs:
            type = attrs['type']
        elif tag in self.tagTypes:
            type = self.tagTypes[tag]
        if type in self.containerTags:
            # I must create a new container object
            if type == 'object':
                new = O(**self.convertAttrs(attrs))
            elif type == 'tuple': new = [] # Tuples become lists
            elif type == 'list': new = []
            elif type == 'dict': new = {}
            elif type == 'file':
                new = UnmarshalledFile()
                if 'name' in attrs:
                    new.name = attrs['name']
                if 'mimeType' in attrs:
                    new.type = attrs['mimeType']
            else: new = O(**self.convertAttrs(attrs))
            # Store the value on the last container, or on the root object
            self.storeValue(tag, new)
            # Push the new object on the container stack
            e.containerStack.append(new)
        else:
            # If we are already parsing a basic type, it means that we were
            # wrong for our diagnostic of the containing element: it was not
            # basic. We will make the assumption that the containing element is
            # then an object.
            if e.currentBasicType:
                # Previous elem was an object: create it on the stack.
                new = O()
                self.storeValue(previousTag, new)
                e.containerStack.append(new)
            e.currentBasicType = type

    def storeValue(self, name, value):
        '''Stores the newly parsed p_value (contained in tag p_name) on the
           current container in environment self.env.'''
        e = self.env
        # Remove namespace prefix when relevant
        if ':' in name: name = name.split(':')[-1]
        # Change the class of the value if relevant
        if (name in self.classes) and isinstance(value, O):
            value.__class__ = self.classes[name]
        # Where must I store this value?
        if not e.containerStack:
            # I store the object at the root of the web
            self.r = value
        else:
            container = e.containerStack[-1]
            if isinstance(container, list):
                container.append(value)
            elif isinstance(container, dict):
                # If the current container is a dict, it means that p_value is
                # a dict entry object named "entry" by convention and having
                # attributes "k" and "v" that store, respectively, the key and
                # the value of the entry. But this object is under construction:
                # at this time, attributes "k" and "v" are not created yet. We
                # will act in m_endElement, when the object will be finalized.
                pass
            elif isinstance(container, UnmarshalledFile):
                val = value or b''
                container.value += val
                container.size += len(val)
            else:
                # Current container is an object
                if hasattr(container, name) and \
                   getattr(container, name):
                    # We have already encountered a sub-object with this name.
                    # Having several sub-objects with the same name, we will
                    # create a list.
                    val = getattr(container, name)
                    if not isinstance(val, list):
                        val = [val, value]
                    else:
                        val.append(value)
                else:
                    val = value
                setattr(container, name, val)

    def characters(self, content):
        e = Parser.characters(self, content)
        if e.currentBasicType:
            if e.currentBasicType == 'base64':
                content = content.strip()
            e.currentContent += content
        elif e.currentTag.name in self.contentAttributes:
            # Store p_content in attribute named according to
            # p_self.contentAttributes on the current object.
            name = self.contentAttributes[e.currentTag.name]
            current = e.containerStack[-1]
            if hasattr(current, name):
                setattr(current, name, getattr(current, name) + content)
            else:
                setattr(current, name, content)

    def endElement(self, tag):
        e = Parser.endElement(self, tag)
        type = e.currentBasicType
        if type:
            value = e.currentContent.strip()
            if not value: value = None
            else:
                # If we have a custom converter for values of this type, use it
                if type in self.conversionFunctions:
                    try:
                        value = self.conversionFunctions[type](value)
                    except Exception as err:
                        raise Exception(CUSTOM_CONVERSION_ERROR % \
                                        (type, value, str(err)))
                # If not, try a standard conversion
                elif type in self.numericTypes:
                    # Manage obsolete type "long"
                    type = 'int' if type == 'long' else type
                    try:
                        value = eval(value)
                    except (SyntaxError, NameError) as e:
                        raise Exception(CONVERSION_ERROR % (type, value))
                    # Convert ints to floats
                    if (type == 'float') and isinstance(value, int):
                        value = float(value)
                    # Check that the value is of the correct type. For instance,
                    # a float value with a comma in it could have been converted
                    # to a tuple instead of a float.
                    if not isinstance(value, eval(type)):
                        raise Exception(CONVERSION_ERROR % (type, value))
                elif type == 'bytes':
                    # Convert the string value to bytes
                    value = value.encode()
                elif type == 'DateTime':
                    value = DateTime(value)
                elif type == 'datetime':
                    value = self.getDatetime(value)
                elif type == 'base64':
                    value = decodebytes(e.currentContent.encode())
            # Store the value on the last container
            self.storeValue(tag, value)
            # Clean the environment
            e.currentBasicType = None
            e.currentContent = ''
        else:
            o = e.containerStack.pop()
            # This element can be a temporary "entry" object representing a dict
            # entry.
            if e.containerStack:
                lastContainer = e.containerStack[-1]
                if isinstance(lastContainer, dict):
                    lastContainer[o.k] = o.v

    # Alias: "unmarshall" > "parse"
    unmarshall = Parser.parse
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This script allows to get information about a given SAP RFC function
   module.'''

# ------------------------------------------------------------------------------
import sys, getpass
from optparse import OptionParser
from appy.shared.sap import Sap, SapError

# ------------------------------------------------------------------------------
WRONG_NG_OF_ARGS = 'Wrong number of arguments.'
ERROR_CODE = 1
P_OPTION = 'The password related to SAP user.'
G_OPTION = 'The name of a SAP group of functions'

# ------------------------------------------------------------------------------
class AskSap:
    '''This script allows to get information about a given RCF function module
       exposed by a distant SAP system.

       usage: %prog [options] host sysnr client user functionName

       "host"         is the server name or IP address where SAP runs.
       "sysnr"        is the SAP system/gateway number (example: 0)
       "client"       is the SAP client number (example: 040)
       "user"         is a valid SAP login
       "sapElement"   is the name of a SAP function (the default) or a given
                      group of functions (if option -g is given). If -g is
                      specified, sapElement can be "_all_" and all functions of
                      all groups are shown.
       Examples
       --------
       1) Retrieve info about the function named "ZFX":
          python asksap.py 127.0.0.1 0 040 USERGDY ZFX -p passwd

       2) Retrieve info about group of functions "Z_API":
          python asksap.py 127.0.0.1 0 040 USERGDY Z_API -p passwd -g

       3) Retrieve info about all functions in all groups:
          python asksap.py 127.0.0.1 0 040 USERGDY _all_ -p passwd -g
    '''
    def manageArgs(self, parser, options, args):
        # Check number of args
        if len(args) != 5:
            print(WRONG_NG_OF_ARGS)
            parser.print_help()
            sys.exit(ERROR_CODE)

    def run(self):
        optParser = OptionParser(usage=AskSap.__doc__)
        optParser.add_option("-p", "--password", action='store', type='string',
            dest='password', default='', help=P_OPTION)
        optParser.add_option("-g", "--group", action='store_true',
            dest='isGroup', default='', help=G_OPTION)
        (options, args) = optParser.parse_args()
        try:
            self.manageArgs(optParser, options, args)
            # Ask the password, if it was not given as an option.
            password = options.password
            if not password:
                password = getpass.getpass('Password for the SAP user: ')
            connectionParams = args[:4] + [password]
            print('Connecting to SAP...')
            sap = Sap(*connectionParams)
            sap.connect()
            print('Connected.')
            sapElement = args[4]
            if options.isGroup:
                # Returns info about the functions available in this group of
                # functions.
                info = sap.getGroupInfo(sapElement)
                prefix = 'Group'
            else:
                # Return info about a given function.
                info = sap.getFunctionInfo(sapElement)
                prefix = 'Function'
            print('%s: %s' % (prefix, sapElement))
            print(info)
            sap.disconnect()
        except SapError, se:
            sys.stderr.write(str(se))
            sys.stderr.write('\n')
            sys.exit(ERROR_CODE)

# ------------------------------------------------------------------------------
if __name__ == '__main__':
    AskSap().run()
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.ui.layout import Layouts
from appy.model.fields import Field

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Custom(Field):
    '''A custom field has the purpose of storing a data structure that is not
       proposed by other available fields.'''

    class Layouts(Layouts):
        '''Info-specific layouts'''
        b = Layouts(edit='f')

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this Custom p_field'''
            return class_.b

    # If your objective is to build a complete custom widget, with specific
    # "edit" and "view" PXs and with a custom data structure, please use a field
    # of type Computed(unfreezable=True).

    # Indeed, the purpose of the Custom field is simpler and aims at storing a
    # data structure that is produced by your code independently of any UI
    # concern. Of course, if you wish, you can define a PX in attribute "view"
    # in order to add UI-visibility.

    # Typically, a Custom field is only "viewable" in the XML layout
    view = edit = cell = search = ''

    def __init__(self, validator=None, multiplicity=(1,1), show='xml',
      page='main', group=None, layouts=None, move=0, readPermission='read',
      writePermission='write', width=None, height=None, maxChars=None,
      colspan=1, master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None, view=None, cell=None,
      edit=None, xml=None, translations=None):
        # Parameter "persist" is not available and is automatically set to True
        Field.__init__(self, None, (0,1), None, None, show, page, group,
          layouts, move, False, True, None, None, False, None, readPermission,
          writePermission, width, height, None, colspan, master, masterValue,
          focus, historized, mapping, generateLabel, label, None, None, None,
          None, True, False, view, cell, edit, xml, translations)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This script allows to walk (and potentially patch) files (content.xml,
   styles.xml...) contained within a given ODF file or within all ODF files
   found in some folder.'''

# ------------------------------------------------------------------------------
import sys, os.path, time
from appy.shared.zip import unzip, zip
from appy.shared.utils import getOsTempFolder, FolderDeleter, executeCommand

# ------------------------------------------------------------------------------
usage = '''Usage: python odfWalk.py [file|folder] yourScript.

 If *file* is given, it is the path to an ODF file (odt or ods). This single
 file will be walked.
 If *folder* is given, we will walk all ODF files found in this folder and
 sub-folders.

 *yourScript* is the path to a Python script that will be run on every walked
 file. It will be called with a single arg containing the absolute path to the
 folder containing the unzipped file content (content.xml, styles.xml...).'''

# ------------------------------------------------------------------------------
class OdfWalk:
    toUnzip = ('.ods', '.odt')
    def __init__(self, fileOrFolder, script):
        self.fileOrFolder = fileOrFolder
        self.script = script
        self.tempFolder = getOsTempFolder()

    def walkFile(self, fileName):
        '''Unzip p_fileName in a temp folder, call self.script, and then re-zip
           the result.'''
        print 'Walking %s...' % fileName
        # Create a temp folder
        name = 'f%f' % time.time()
        tempFolder = os.path.join(self.tempFolder, name)
        os.mkdir(tempFolder)
        # Unzip the file in it
        unzip(fileName, tempFolder)
        # Call self.script
        py = sys.executable or 'python'
        cmd = [py, self.script, tempFolder]
        print '  Running %s...' % cmd,
        out, err = executeCommand(cmd)
        # Re-zip the result
        zip(fileName, tempFolder, odf=True)
        FolderDeleter.delete(tempFolder)
        print 'done.'

    def run(self):
        if os.path.isfile(self.fileOrFolder):
            self.walkFile(self.fileOrFolder)
        elif os.path.isdir(self.fileOrFolder):
            # Walk all files found in this folder
            for dir, dirnames, filenames in os.walk(self.fileOrFolder):
                for name in filenames:
                    if os.path.splitext(name)[1] in self.toUnzip:
                        self.walkFile(os.path.join(dir, name))
        else:
            print('%s does not exist.' % self.fileOrFolder)

# ------------------------------------------------------------------------------
if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(usage)
        sys.exit()
    # Warn the user.
    print 'All the files in %s will be modified. ' \
          'Are you sure? [y/N] ' % sys.argv[1],
    response = sys.stdin.readline().strip().lower()
    if response == 'y':
        OdfWalk(sys.argv[1], sys.argv[2]).run()
    else:
        print 'Canceled.'
# ------------------------------------------------------------------------------



'''Initiators for searches whose results are shown in popups'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

# Abstract base class for any initiator  - - - - - - - - - - - - - - - - - - - -
class Initiator:
    '''When a search is rendered in a popup, the "initiator", in the main page,
       can be:
       * (a) some object, in view or edit mode, displaying a given Ref field
             for which the popup is used to select one or more objects to link;
       * (b) some class for which we must create an instance from a template;
             the popup is used to select such a template object.

       This class is the abstract class for 2 concrete initiator classes:
       RefInitiator (for case a) and TemplateInitiator (for case b).
    '''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class RefInitiator(Initiator):
    def __init__(self, o, field, fieldName, mode):
        # The initiator object
        self.o = o
        # The initiator field
        self.field = field
        # As usual, the field name can be different from field.name if it is a
        # sub-field within an outer field.
        self.fieldName = fieldName
        # The mode can be:
        # - "repl" if the objects selected in the popup will replace already
        #          tied objects;
        # - "add"  if those objects will be added to the already tied ones.
        self.mode = mode
        # "hook" is the ID of the initiator field's XHTML tag
        self.hook = '%d_%s' % (o.iid, fieldName)
        # The root Ajax hook ID in the popup
        self.popupHook = '%s_popup' % self.hook

    def showCheckboxes(self):
        '''We must show object checkboxes if self.field is multivalued: indeed,
           in this case, several objects can be selected in the popup.'''
        return self.field.isMultiValued()

    def jsSelectOne(self, q, cbId):
        '''Generates the Javascript code to execute when a single object is
           selected in the popup.'''
        return 'onSelectObject(%s,%s,%s)' % \
               (q(cbId), q(self.hook), q(self.o.iid))

    def jsSelectMany(self, q, sortKey, sortOrder, filters):
        '''Generates the Javascript code to execute when several objects are
           selected in the popup.'''
        return 'onSelectObjects(%s,%s,%s,%s,null,%s,%s,%s)' % \
          (q(self.popupHook), q(self.hook), q(self.o.url), q(self.mode), \
           q(sortKey or ''), q(sortOrder), q(filters))

    def getAjaxParams(self):
        '''Get initiator-specific parameters for retriggering the Ajax
           request for refreshing objects in the popup.'''
        return

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TemplateInitiator(Initiator):
    MANY_ERROR = 'Cannot select several objects from a template initiator.'

    def __init__(self, className, formName, insert, sourceField):
        # The class from which we must create an instance based on a template
        # that we will choose in the popup. Indeed, the instance to create may
        # be from a different class that the instances shown in the popup.
        self.className = className
        # The name of the form that will be submitted for creating the object
        # once a template will have been selected in the popup.
        self.formName = formName
        # The root Ajax hook ID in the popup
        self.popupHook = '%s_popup' % className
        # If the object to create must be inserted at a given place in a Ref
        # field, this can be specified in p_insert.
        self.insert = insert or ''
        # The source field
        self.sourceField = sourceField

    def showCheckboxes(self):
        '''We must hide object checkboxes: only one template object can be
           selected.'''
        return

    def jsSelectOne(self, q, cbId):
        '''Generates the Javascript code to execute when a single object is
           selected in the popup.'''
        return 'onSelectTemplateObject(%s,%s,%s)' % \
               (q(cbId), q(self.formName), q(self.insert))

    def jsSelectMany(self, q, sortKey, sortOrder, filters):
        raise Exception(self.MANY_ERROR)

    def getAjaxParams(self):
        r = {'fromClass': self.className, 'formName': self.formName}
        if self.insert:
            r['insert'] = self.insert
        if self.sourceField:
            r['sourceField'] = self.sourceField
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class SearchInitiator:
    '''Initiator in use when an object is created from a search'''

    def __init__(self, tool, req, info):
        self.tool = tool
        self.req = req
        # Extract the initiator object and field from p_info, parsed from a
        # "nav" key in the request.
        self.info = info.split('.')
        # Get info about the field from which the object creation was triggered
        if ',' in self.info[1]:
            id, name, layout = self.info[1].split(',')
            self.o = tool.getObject(id)
            self.field = self.o.getField(name)
        else:
            # There is no effective initiator
            self.o = self.field = None
        self.backFromPopupUrl = None

    # Conform to the Initiator API
    def checkAllowed(self, log=False): pass
    def updateParameters(self, params): pass
    def manage(self, new): pass
    def goBack(self): return 'view'

    def getUrl(self):
        '''Returns the URL for going back to the initiator object, on the page
           showing self.field.'''
        return self.o.getUrl(sub='view', page=self.field.pageName, nav='no')

    def getBackUrl(self, o): return self.getUrl()

    def getNavInfo(self, new):
        '''Reconstitute the navInfo for navigating from p_new'''
        # It is impossible to know at what position p_new will come within the
        # originating search. By specifying the total number of elements being
        # 0, it deactivates sibling navigation and keeps only the icon to go
        # back to the complete search.
        return 'search.%s.%s.0.0' % (self.info[0], self.info[1])

    def isComplete(self):
        '''The initiator is considered to be complete if "obj" is not empty'''
        return self.o
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# ------------------------------------------------------------------------------
from appy.pod.xhtml import tags, visitors
from appy.shared.xml_parser import XmlEnvironment, XmlParser, Escape

# ------------------------------------------------------------------------------
CHUNK_NOT_WITH_FILE = 'Cannot parse a file when chunk=True, only a string.'

# ------------------------------------------------------------------------------
class XhtmlEnvironment(XmlEnvironment):
    def __init__(self):
        # Call the base constructor
        XmlEnvironment.__init__(self)
        # The parsing result will store a root Tag instance
        self.r = None
        # The currently walked tag (a tags.Tag or sub-class' instance)
        self.current = None
        # The current content (a tags.Content instance)
        self.content = None
        # Tags encountered so far, by type
        self.tags = {}
        # The main parser (see below)
        self.parser = None

    def addTag(self, tag):
        '''A new p_tag has been encountered. Add it in the environment's local
           data structures.'''
        # Store it as the current tag
        if not self.r:
            # It is the root tag
            self.r = tag
            isRoot = True
        else:
            isRoot = False
        self.current = tag
        # Store it in self.tags if walkable. Do not store the root tag: it is a
        # tag that was artificially added to ensure there is a single root tag.
        klass = tag.__class__
        if not isRoot and klass.walkable:
            name = klass.__name__
            if name in self.tags:
                self.tags[name].append(tag)
            else:
                self.tags[name] = [tag]

# ------------------------------------------------------------------------------
class XhtmlParser(XmlParser):
    '''Creates a tree of Tag elements from a chunk of XHTML code as a string'''

    def __init__(self, env=None, caller=None, chunk=False, encoded=False,
                 compress=True, normalizeTables=True, optimizeTables=False,
                 keepWithNext=0, removeTrailingParas=None):
        # Define a default environment if p_env is None
        env = env or XhtmlEnvironment()
        # Call the base constructor
        XmlParser.__init__(self, env, caller)
        # If the XHTML to parse is not a complete, single-root-tag XML, but,
        # instead, a chunk of XHTML potentially containing several tags at the
        # root level, p_chunk is True and the XHTML chunk will be surrounded by
        # a faxe "x" tag.
        self.chunk = chunk
        # The parser will output UTF-8 in a unicode string. If you need it to be
        # encoded as a str, specify p_encoded being True.
        self.encoded = encoded
        # By default, "compress" is True: the parser will remove ignorable
        # whitespace to produce a shorter output. If it is a requirement to keep
        # the general formatting of the input XHTML, set p_compress to False.
        self.compress = compress
        # Define the visitors that will walk the tree of tags we will build
        activated = []
        if normalizeTables:
            activated.append(visitors.TablesNormalizer())
        if optimizeTables:
            activated.append(visitors.TablesOptimizer())
        if keepWithNext:
            activated.append(visitors.KeepWithNext(keepWithNext))
        if removeTrailingParas is not None:
            cleaner = visitors.Cleaner(removeTrailingParas=removeTrailingParas)
            activated.append(cleaner)
        self.visitors = activated
        # This flag is True if the input XHTML has been modified by at least one
        # visitor.
        self.updated = False

    def parse(self, xml, source='string'):
        '''Surrounds p_xml with a fake root tag if p_self.chunk is True, before
           calling the base method.'''
        # Surrounding with a fake tag is only possible if p_source is "string"
        chunk = self.chunk
        if chunk:
            if source == 'string':
                xml = '<x>%s</x>' % xml
            else:
                raise Exception(CHUNK_NOT_WITH_FILE)
        # Call the base method
        return XmlParser.parse(self, xml, source=source)

    def startElement(self, elem, attrs):
        '''Manages the start of a tag'''
        env = self.env
        # Clean current content
        env.content = None
        # Create the corresponding Tag instance
        tag = tags.get(elem)(elem, attrs, parent=env.current)
        env.addTag(tag)

    def endElement(self, elem):
        '''Manages the end of a tag'''
        env = self.env
        # Clean current content
        env.content = None
        # We are back to p_elem's parent
        env.current = env.current.parent

    def characters(self, text):
        '''Manages text encountered within the current tag'''
        env = self.env
        # Ignore p_text if directly contained in a tag that is not supposed to
        # directly contain content. In that case, p_text is probably ignorable
        # whitespace. Avoid doing this if p_self.compress is False.
        if env.current.structural and self.compress: return
        # Xhtml-escape p_text and store it on the environment
        text = Escape.xml(text)
        if env.content:
            env.content.text += text
        else:
            # Create a Content instance to store p_text
            content = env.content = tags.Content(text)
            # Add content as child of the current tag
            tag = env.current
            if tag.children:
                tag.children.append(content)
            else:
                tag.children = [content]

    def endDocument(self):
        '''Converts p_self.env.r into a chunk of ODF code and store it in
           p_self.res.'''
        env = self.env
        # Accept the visitors
        for visitor in self.visitors:
            updated = visitor.visit(env)
            self.updated = self.updated or updated
        self.res = env.r.asXhtml()
        if self.chunk:
            # Remove the base tag
            self.res = self.res[3:-4]
        if self.encoded:
            self.res = self.res.encode('utf-8')
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>This para should get style Signature.</p>
<ul>
 <li><p>This one shouldn't</p></li>
</ul>
<table>
 <tr><td><p>This one shouldn't, too.</p></td></tr>
</table>
<div><div class="Signature">Match</div></div>
'''



# -*- coding: utf-8 -*-
xhtmlInput = '''<strong>A<strong>B</strong></strong>
  <p>Text "Hello" in the following span tag will be applied an
  inline style: <span class="Variable">Hello</span> end.</p>
  <p>And here, style <span class="TELE">Teletype</span> will be applied</p>
  <p><br/>Here, underline and bold were not noth applied</p>
  <p><u><strong>Text :</strong></u></p>
  <p><b><u>Text :</u></b></p>
  <p><b><u>Text :</u> continued</b></p>
  <p>In the text below, some chars were deleted:</p>
  <p>My text,<span> text with space before</span> text
     text,<span> text with space before and after </span>and text.</p>
  <p><u><s><strong>Bold strike underline</strong></s></u></p>
  <p><b><i><s>Bold strike italic</s></i>;</b></p>
  <p style="font-size: small">Small</p>
  <p><em>Tralala<br />
   <strong>- 1 ;<br />
   - 2 ;<br />
    - 3 ;<br />
   - 4.</strong></em>
  </p>
  <!-- The space before the star was removed -->
  <p><em><em>Texte</em></em><b><u> *texte</u></b></p>
  <p style="text-align:justify"><em><em>Texte </em></em><em><em>texte texte</em></em><em><em> texte </em></em><em><em>texte</em></em><em><em> texte.</em></em></p>
  <!-- The "strike" was not applied throughout the whole paragraph -->
  <p><s><u>test </u><u>...</u>must be strike here <u>.</u></s></p>
  <p><span style="color:rgba(234, 44, 33, 9)">Perpette les foins</span></p>
  <p><span style="color:rgba(234, 44, 33, 0.6)">Perpette les foins</span></p>
  <table style="width:17cm" cellspacing="1" cellpadding="1" border="1">
   <tbody>
    <tr>
     <td style="background-color:rgb(0, 200, 100)">Use of <b>"colspan"</b> and rgb()</td>
     <td style="background-color:rgba(0, 0, 255, 0.5)">Use of rgba(), crashes wit Appy &lt;= 0.9.10...</td>
    </tr>
   </tbody>
  </table>
  <p style="font-size: 85%">85% : text</p>
  <p style="font-size: 100%">100% : text</p>
  <p style="font-size: 115%">115% : text</p>
'''



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>Hello.</p>
<h2>Heading One</h2>
Blabla.<br />
<h3>SubHeading then.</h3>
Another blabla.<br /><br /><br />
<p style="font-weight:bold">Hello CentreCentre</p>
<div>In a div</div>
'''
# I need a class
class D:
    def getAt1(self):
        return xhtmlInput
dummy = D()

xhtmlInput2='''
<p style="font-style:italic">Mapped to style_1.</p>
'''

xhtmlInput3='''
<p>The text within cells must be in style Blue</p>
<table><tr><td><p>Must be Blue</p></td><td><p>Here too</p></td></tr></table>
'''



'''Meta-class for a Appy workflow'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import collections

from appy.px import Px
from appy.utils import No
from appy.model import Model
from appy.model.meta import Meta
from appy.model.workflow.state import State
from appy.model.workflow.transition import Transition
from appy.model.fields.phase import Page as FieldPage

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
MULTIPLE_INITIAL_STATES = 'In workflow "%s", states "%s" and "%s" are both ' \
  'defined as being initial.'
NO_INITIAL_STATE = 'Workflow "%s" does not define any initial state.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Workflow(Meta):
    '''Represents a Appy worflow'''

    # Attributes added on Appy workflows
    attributes = {'states': State, 'transitions': Transition}

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Model-construction-time methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # The following set of methods allow to build the model when making an app
    # or when starting the server.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def __init__(self, class_, isBase=False, appOnly=False):
        Meta.__init__(self, class_, appOnly)
        # p_class_, representing a workflow, may be of 2 types:
        # ----------------------------------------------------------------------
        # "base" | A standard workflow from package appy.model.workflow.standard
        # ----------------------------------------------------------------------
        # "app"  | An original workflow defined in the app.
        # ----------------------------------------------------------------------
        # Unlike classes, standard workflows can't be overridden
        self.type = 'base' if isBase else 'app'
        # The concrete class for a workflow is p_class_ directly. Indeed,
        # workflows are never instantiated: we don't need to create a specific
        # concrete class like for classes.
        self.concrete = class_
        # Read states and transitions defined on this workflow
        self.readStatesAndTransitions()
        # At run-time, we need a single instance of this workflow, in order to
        # execute workflow-related methods defined as conditions and actions.
        # We create here this "prototypical" instance.
        if not appOnly: self.proto = class_()

    def readStatesAndTransitions(self):
        '''Create attributes related to states and transitions for this
           worflow.'''
        # ----------------------------------------------------------------------
        # The following attributes will be created.
        # "OD" is a shorthand for "OrderedDict".
        # ----------------------------------------------------------------------
        # Name              | Type       | Description
        # ----------------------------------------------------------------------
        # states            | OD         | The ordered dict of workflow states
        #                   |            | as appy.model.workflow.state.State
        #                   |            | instances.
        # ----------------------------------------------------------------------
        # transitions       | OD         | The ordered dict of workflow
        #                   |            | transitions as appy.model.workflow.\
        #                   |            | transition.Transition instances.
        # ----------------------------------------------------------------------
        # initialState      | State      | The workflow's initial state, that
        #                   |            | is aslo present among "states".
        # ----------------------------------------------------------------------
        # initialTransition | Transition | A virtual transition, named "_init_",
        #                   |            | added while loading the model, not
        #                   |            | present among "transitions" and that
        #                   |            | is triggered when the object is
        #                   |            | created, adding an initial entry in
        #                   |            | the object's history.
        # ----------------------------------------------------------------------
        states = collections.OrderedDict()
        transitions = collections.OrderedDict()
        initial = None
        Err = Model.Error
        for name, value in self.concrete.__dict__.items():
            if name.startswith('__'): continue
            if isinstance(value, State):
                # A state was found. Ensure its name is acceptable.
                self.checkAttributeName(name)
                # Late-initialize it
                value.init(self, name)
                # Store it
                states[name] = value
                if value.initial:
                    # There can be at most one initial state
                    if initial:
                        raise Err(MULTIPLE_INITIAL_STATES % \
                                  (self.name, initial.name, name))
                    initial = value
            elif isinstance(value, Transition):
                # A transition was found. Ensure its name is acceptable.
                self.checkAttributeName(name)
                # Late-initialize it
                value.init(self, name)
                # Store it
                transitions[name] = value
        # At least one initial state is required
        if not initial: raise Err(NO_INITIAL_STATE % self.name)
        self.states = states
        self.transitions = transitions
        self.initialState = initial
        # Create a Transition instance representing the initial transition
        self.initialTransition = tr = Transition((initial, initial))
        tr.init(self, '_init_')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Run-time methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # The following set of methods allow to compute, at run-time,
    # workflow-dependent elements.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def getRolesFor(self, o, permission):
        '''Gets the roles that are currently granted p_permission on this
           p_o(bject). r_ is a list of role names.'''
        state = self.states[o.state]
        return state.getRolesFor(permission)

    def getTransitions(self, o, includeFake=True, includeNotShowable=False,
                       grouped=True):
        '''Return transitions that the current user can trigger from the user
           interface, as a list of UiTransition instances.'''
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #    includeFake     | If True, it retrieves transitions that the user
        #                    | can't trigger, but for which he needs to know for
        #                    | what reason he can't trigger it.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # includeNotShowable | If True, it includes transitions for which
        #                    | show=False. Indeed, because "showability" is only
        #                    | a UI concern, and not a security concern, in some
        #                    | cases it has sense to set this parameter to True,
        #                    | because those transitions are triggerable from a
        #                    | security point of view.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #      grouped       | If True, transitions are grouped according to
        #                    | their "group" attribute, in a similar way to
        #                    | fields or searches.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        r = []
        groups = {} # The already encountered groups of transitions
        page = FieldPage('workflow')
        currentState = self.states[o.state]
        # Walk workflow transitions
        for transition in self.transitions.values():
            # Filter transitions that do not have currentState as start state
            if not transition.hasState(currentState, True): continue
            # Check if the transition can be triggered
            mayTrigger = transition.isTriggerable(o)
            # Compute the condition that will lead to including or not this
            # transition.
            includeIt = mayTrigger if not includeFake \
                        else mayTrigger or isinstance(mayTrigger, No) 
            if not includeNotShowable:
                includeIt = includeIt and transition.isShowable(o)
            if not includeIt: continue
            # Create the UiTransition instance
            ui = transition.ui(o, mayTrigger)
            # Add the transition into the result
            if not transition.group or not grouped:
                r.append(ui)
            else:
                # Insert the UiGroup instance corresponding to transition.group
                uiGroup = transition.group.insertInto(r, groups, page,
                                           o.class_.name, content='transitions')
                uiGroup.addElement(ui)
        return r

    def listStates(self, o):
        '''Returns the list of (not-isolated) translated names for states in
           this workflow.'''
        return [(state.name, o.translate(state.labelId)) \
                for state in self.states.values() if not state.isIsolated()]

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # PXs
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Displays transitions as icons or buttons for an object (variable "o")
    pxTransitions = Px('''
     <x var="transitions=transitions|workflow.getTransitions(o)"
        if="transitions"
            var2="formId='trigger_%d' % o.iid;
                  backHook=backHook|None">
      <form id=":formId" method="post" class="inline" data-baseurl=":o.url">
       <input type="hidden" name="popup" value=":popup"/>
       <input type="hidden" name="nav" value=":req.nav or 'no'"/>
       <input type="hidden" name="page" value=":req.page or 'main'"/>
       <!-- Input field for storing the comment coming from the popup -->
       <textarea name="popupComment" cols="30" rows="3"
                 style="display:none"></textarea>
      </form>

      <!-- Render a transition or a group of transitions. "inline" will be
           overridden if this transition is rendered inside a flexbox. -->
      <div for="transition in transitions" class="inline"
           var2="trGroup=transition">:transition.px</div>
     </x>''')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Sort results produced by a catalog'''

# Code inspired by Zope's catalog

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from BTrees.IIBTree import intersection

from appy.database.lazy import LazyCat, LazyValues

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Sorter:
    '''Sorts results as produced by a catalog'''

    # The sorter' input and output data structures are as follows.
    # --------------------------------------------------------------------------
    # input   | A result set of object IIDs as computed by a catalog, as an
    #         | instance of IISet.
    # --------------------------------------------------------------------------
    # output  | A sorted list of object IIDs, in the form of a Lazy data
    #         | structure (see appy/database/lazy.py). The sorter produces an
    #         | intermediary data structure that is more complex than a list.
    #         | Instead of directly and completely "flattening" this data
    #         | structure into a list of IIDs, we wrap it, for performance
    #         | reasons, into a Lazy data structure, that behaves like a list
    #         | and manipulates, behind the scenes, the more complex internal
    #         | structure.
    # --------------------------------------------------------------------------

    def __init__(self, sortIndex, rs, reverse):
        # The index to use for sorting
        self.sortIndex = sortIndex
        # Must we reverse results ?
        self.reverse = reverse
        # The (unsorted) result set
        self.rs = rs

    def iterateByIndex(self, r):
        '''Fills list p_r with sorted entries from p_self.rs, by iterating over
           p_self.sortIndex.'''
        rs = self.rs
        length = 0
        for value, iids in self.sortIndex.byValue.items():
            # Object "iids" share the same index "value". Intersect each set of
            # "iids" with the result set, and produce a sorted list of these
            # intersections.
            subset = intersection(rs, iids)
            if subset:
                subset = subset.keys()
                r.append((value, subset))
                length += len(subset)
        return length

    def iterateByResultSet(self, r):
        '''Fills list p_r with sorted entries frol p_self.rs by iterating over
           p_self.rs.'''
        rs = self.rs
        index = self.sortIndex
        for iid in rs:
            value = index.byObject.get(iid)
            # p_value can be None if the object is not in the sort index
            r.append((value, iid))
        return len(r)

    def run(self):
        '''Returns a Lazy data structure containing object IIDs found in
           p_self.rs, sorted by p_self.sortIndex.'''
        r = []
        # Choose what method to use for sorting p_self.rs. It is more performant
        # to iterate over the index if shorter than the result set.
        rsLength = len(self.rs)
        indexLength = len(self.sortIndex.byValue)
        algo = 'ByIndex' if rsLength > (indexLength * (rsLength / 100 + 1)) \
                         else 'ByResultSet'
        # Use the chosen algorithm to produce an intermediate result in p_r
        length = getattr(self, 'iterate%s' % algo)(r)
        # Sort p_r. Reverse the sort order when appropriate.
        r.sort(reverse=self.reverse)
        # Return the list as a "lazy" data structure (see appy.database.lazy.py)
        if algo == 'ByIndex':
            # Sort results are in the form of a list of sub-lists
            r = LazyCat(LazyValues(r), length=length)
        else:
            # Sort results are in the form of a simple list, but whose
            # sub-elements are (value, iid) tuples instead of simple IIDs.
            r = LazyValues(r)
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-

'''This folder contains copies of external, "authentic" data, stored as text
   files, like ISO 3166-1 country codes. In this package, corresponding Python
   classes are available for accessing the data in the text files.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path

# Languages names in their own language  - - - - - - - - - - - - - - - - - - - - 
nativeNames = {
  'aa' : 'магIарул мацI',
  'ab' : 'бызшәа',
  'af' : 'Afrikaans',
  'am' : 'አማርኛ',
  'ar' : 'العربية',
  'as' : 'অসমিয়া',
  'ay' : 'Aymara',
  'az' : 'Azəri Türkçəsi',
  'ba' : 'Bashkir',
  'be' : 'Беларускі',
  'bg' : 'Български',
  'bh' : 'Bihari',
  'bi' : 'Bislama',
  'bn' : 'বাংলা',
  'bo' : 'བོད་སྐད་',
  'bs' : 'Bosanski',
  'br' : 'Brezhoneg',
  'ca' : 'Català',
  'ch' : 'Chamoru',
  'co' : 'Corsu',
  'cs' : 'Čeština',
  'cy' : 'Cymraeg',
  'da' : 'Dansk',
  'de' : 'Deutsch',
  'dz' : 'རྫོང་ཁ',
  'el' : 'Ελληνικά',
  'en' : 'English',
  'eo' : 'Esperanto',
  'es' : 'Español',
  'et' : 'Eesti',
  'eu' : 'Euskara',
  'fa' : 'فارسی',
  'fi' : 'Suomi',
  'fj' : 'Fiji',
  'fo' : 'Føroyska',
  'fr' : 'Français',
  'fy' : 'Frysk',
  'ga' : 'Gaeilge',
  'gd' : 'Gàidhlig',
  'gl' : 'Galego',
  'gn' : 'Guarani',
  'gu' : 'ગુજરાતી',
  'gv' : 'Gaelg',
  'ha' : 'هَوُس',
  'he' : 'עברית',
  'hi' : 'हिंदी',
  'hr' : 'Hrvatski',
  'hu' : 'Magyar',
  'hy' : 'Հայերէն',
  'ia' : 'Interlingua',
  'id' : 'Bahasa Indonesia',
  'ie' : 'Interlingue',
  'ik' : 'Inupiak',
  'is' : 'Íslenska',
  'it' : 'Italiano',
  'iu' : 'ᐃᓄᒃᑎᑐᑦ',
  'ja' : '日本語',
  'jbo': 'lojban',
  'jw' : 'Basa Jawi',
  'ka' : 'ქართული',
  'kk' : 'ﻗﺎﺯﺍﻗﺸﺎ',
  'kl' : 'Greenlandic',
  'km' : 'ខ្មែរ',
  'kn' : 'ಕನ್ನಡ',
  'ko' : '한국어',
  'ks' : 'काऽशुर',
  'ku' : 'Kurdí',
  'kw' : 'Kernewek',
  'ky' : 'Кыргыз',
  'la' : 'Latin',
  'lb' : 'Lëtzebuergesch',
  'li' : 'Limburgs',
  'ln' : 'Lingala',
  'lo' : 'ພາສາລາວ',
  'lt' : 'Lietuviskai',
  'lv' : 'Latviešu',
  'mg' : 'Malagasy',
  'mi' : 'Maori',
  'mk' : 'Македонски',
  'ml' : 'മലയാളം',
  'mn' : 'Монгол',
  'mo' : 'Moldavian',
  'mr' : 'मराठी',
  'ms' : 'Bahasa Melayu',
  'mt' : 'Malti',
  'my' : 'Burmese',
  'na' : 'Nauru',
  'ne' : 'नेपाली',
  'nl' : 'Nederlands',
  'no' : 'Norsk',
  'nn' : 'Nynorsk',
  'oc' : 'Languedoc',
  'om' : 'Oromo',
  'or' : 'ଓଡ଼ିଆ',
  'pa' : 'ਪੰਜਾਬੀ',
  'pl' : 'Polski',
  'ps' : 'پښتو',
  'pt' : 'Português',
  'qu' : 'Quechua',
  'rm' : 'Rumantsch',
  'rn' : 'Kirundi',
  'ro' : 'Română',
  'ru' : 'Русский',
  'rw' : 'Kiyarwanda',
  'sa' : 'संस्कृत',
  'sd' : 'Sindhi',
  'se' : 'Northern Sámi',
  'sg' : 'Sangho',
  'sh' : 'Serbo-Croatian',
  'si' : 'Singhalese',
  'sk' : 'Slovenčina',
  'sl' : 'Slovenščina',
  'sm' : 'Samoan',
  'sn' : 'Shona',
  'so' : 'Somali',
  'sq' : 'Shqip',
  'sr' : 'српски',
  'ss' : 'Siswati',
  'st' : 'Sesotho',
  'su' : 'Sudanese',
  'sv' : 'Svenska',
  'sw' : 'Kiswahili',
  'ta' : 'தமிழ',
  'te' : 'తెలుగు',
  'tg' : 'Тоҷики',
  'th' : 'ไทย',
  'ti' : 'ትግርኛ',
  'tk' : 'түркmенче',
  'tl' : 'Tagalog',
  'tn' : 'Setswana',
  'to' : 'Lea faka-Tonga',
  'tr' : 'Türkçe',
  'ts' : 'Tsonga',
  'tt' : 'татарча',
  'tw' : 'Twi',
  'ug' : 'Uigur',
  'uk' : 'Українська',
  'ur' : 'اردو',
  'uz' : 'Ўзбекча',
  'vi' : 'Tiếng Việt',
  'vo' : 'Volapük',
  'wa' : 'Walon',
  'wo' : 'Wolof',
  'xh' : 'isiXhosa',
  'yi' : 'ײִדיש',
  'yo' : 'Yorùbá',
  'za' : 'Zhuang',
  'zh' : '中文',
  'zu' : 'isiZulu'
}
# List of languages having direction right-to-left (RTL) - - - - - - - - - - - -
rtlLanguages = ('ar', 'he', 'fa')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Languages:
    '''This class gives access to the language codes and names as standardized
       by ISO-639. The file has been downloaded in July 2009 from
       http://www.loc.gov/standards/iso639-2/ascii_8bits.html (UTF-8 version)'''

    def __init__(self):
        self.path = Path(__file__).parent / 'Languages.Iso639-2'
        self.languageCodes = []
        # Names of languages in English
        self.languageNames = []
        # Names of languages in their language. It is not part of ISO 639.2 and
        # is taken from dict languageNames above.
        self.nativeNames = []
        self.parseFile()

    def parseFile(self):
        '''Parses the language codes and names in the ISO file and puts them in
           self.languageCodes, self.languageNames and self.nativeNames.'''
        with self.path.open(encoding='utf-8') as f:
            for line in f:
                if line.strip():
                    lineElems = line.split('|')
                    if lineElems[2].strip():
                        # I take only those that have a 2-chars ISO-639-1 code
                        self.languageCodes.append(lineElems[2])
                        self.languageNames.append(lineElems[3])
                        if lineElems[2] in nativeNames:
                            self.nativeNames.append(nativeNames[lineElems[2]])
                        else:
                            # Put the english name nevertheless.
                            self.nativeNames.append(lineElems[3])

    def exists(self, code):
        '''Is p_code a valid 2-digits language code ?'''
        return code in self.languageCodes

    def get(self, code):
        '''Returns information about the language whose code is p_code'''
        try:
            iCode = self.languageCodes.index(code)
            return self.languageCodes[iCode], self.languageNames[iCode], \
                   self.nativeNames[iCode]
        except ValueError:
            return None, None, None

    def __repr__(self):
        i = -1
        res = ''
        for languageCode in self.languageCodes:
            i += 1
            res += 'Language: ' + languageCode + ' - ' + self.languageNames[i]
            res += '\n'
        return res

# Country codes ISO 3166-1 - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Countries:
    '''This class gives access to the country codes and names as standardized by
       ISO 3166-1. The file has been downloaded in March 2011 from
       http://www.iso.org/iso/country_codes/iso_3166_code_lists.htm
       (first line has been removed).'''
    # The single Countries instance will be stored here
    instance = None

    @classmethod
    def get(class_):
        '''Returns the single existing instance of this class in
           p_class_.instance. If it does noe exist yet, it is created.'''
        r = class_.instance
        if not r:
            class_.instance = r = class_()
        return r

    def __init__(self):
        # This file has been downloaded from
        # http://www.iso.org/iso/country_codes.htm and converted to utf-8.
        self.path = Path(__file__).parent / 'Countries.Iso3166-1'
        self.countryCodes = []
        # Names of countries in English
        self.countryNames = []
        self.parseFile()

    def parseFile(self):
        with self.path.open() as f:
            for line in f:
                if line.strip():
                    name, code = line.split(';')
                    self.countryCodes.append(code.strip())
                    self.countryNames.append(name.strip())

    def exists(self, code):
        '''Is p_code a valid 2-digits country code?'''
        return code in self.countryCodes
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
xhtmlInput='''
<!-- "margin-left" property on paragraphs -->
<p>Normal paragraph with a specia char\n\t\v\x0cUn montant de \x0b\v145,00 € TVAC</p>
<p style="margin-left: 25px">25px margin left</p>
<p style="margin-left: 4cm">4cm margin left</p>
<p style="margin-left:-1cm">Text with a negative left margin</p>
<p style="margin-left:-.75pt">Mon texte avec retrait de margin-left:-.75pt;</p>

<!-- Text alignment within table cells (ckeditor-generated) -->
<table border="1" cellpadding="1" cellspacing="1" style="width:500px">
 <tbody>
  <tr>
   <td>Left-aligned text</td>
   <td style="text-align: center;">Centered text</td>
  </tr>
  <tr>
   <td style="text-align: right;">Right-aligned text</td>
   <td></td>
  </tr>
 </tbody>
</table>

<!-- Text alignment within paragraphs and divs -->
<p style="text-align: left">Left-aligned text</p>
<p style="text-align: right; margin-right: 30px">Right-aligned + margin rigth</p>
<p style="text-align: justify; margin-left: 2cm; margin-right: 2cm">Justified ext between margins Justified ext between margins
Justified ext between margins Justified ext between margins Justified ext between margins Justified ext between margins Justified ext between margins
justified ext between margins Justified ext between margins Justified ext between margins Justified ext between margins</p>
<div style="text-align: center; margin-top: 30px">Margin top + center in a div</div>
<p style="margin-left: 4cm">Second with 4cm margin left, to check if a single style will be generated</p>

<!-- Transfering styles from deleted inner-p tags -->
<table>
  <tr>
    <td><p style="text-align:right;">Right-aligned text, style applied on a p inside a td</p></td>
    <td><p style="text-align:center;">Centered text, style applied on a p inside a td</p></td>
    <td><p class="my_class">Text with a CSS class defined on the inner p</p></td>
  </tr>
</table>
<ul>
  <li><p style="font-style: italic; font-weight: bold;">Attribute "style" bold+italic</p></li>
  <li><p style="text-decoration: underline">Underline</p></li>
  <li><p class="Signature">"class" attribute using style "Signature"</p></li>
</ul>
<table border="1">
  <tbody>
    <tr>
      <td><p style="text-decoration: overline">Overlined text</p>
          <p style="font-weight: bold">This will be bold, style is not ignored anymore.</p></td>
      <td style="text-decoration: line-through">Line-through</td>
    </tr>
  </tbody>
</table>
<ul>
  <li style="font-style: italic; text-decoration: line-through;">italique barré;</li>
</ul>
<table>
 <tr><td style="font-style: italic">TD italique</td></tr>
</table>
<ol start="1" style="list-style-type: decimal; margin-left: 40px;">
 <li>
  <p>Item with left indent</p>
 </li>
</ol>
<p style="line-height: normal">Paragraph with a normal line-height</p>
<p style="line-height: 1.5em">Paragraph with a line-height of 1.5em.</p>
'''



# ------------------------------------------------------------------------------
from appy.gen import WorkflowOwner
from appy.gen.wrappers import AbstractWrapper

# ------------------------------------------------------------------------------
class GroupWrapper(AbstractWrapper):
    workflow = WorkflowOwner

    def showLogin(self):
        '''When must we show the login field ?'''
        if self.o.isTemporary(): return 'edit'
        return ('view', 'result')

    def showGroups(self):
        '''Only the admin can view or edit roles'''
        return self.user.has_role('Manager')

    def validateLogin(self, login):
        '''Is this p_login valid ?'''
        return True

    def getGrantableRoles(self):
        '''Returns the list of roles that the admin can grant to a user'''
        res = []
        for role in self.o.getProductConfig().grantableRoles:
            res.append( (role, self.translate('role_%s' % role)) )
        return res

    def validate(self, new, errors):
        '''Inter-field validation'''
        return self._callCustom('validate', new, errors)

    def getMailRecipients(self):
        '''Gets the list of mail recipients for every group member'''
        r = []
        for user in self.users:
            recipient = user.getMailRecipient()
            if recipient:
                r.append(recipient)
        return r

    def onEdit(self, created):
        # If the group was created by anon|system, anon|system can't stay Owner
        self.deleteLocalRole(('anon', 'system'))
        # Ensure the group has a title
        if self.isEmpty('title'): self.title = self.login
        return self._callCustom('onEdit', created)
# ------------------------------------------------------------------------------



'''An index for Ref fields'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.database.indexes import Index

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class RefIndex(Index):
    '''Index for a Ref field'''

    @classmethod
    def toIndexed(class_, value, field):
        '''Converts p_value, a list of tied objects, into their internal index
           representation = a list of their IIDs.'''
        # If there is no value at all, nothing must be indexed
        if value is None:
            return
        elif isinstance(value, int):
            # Already an object IID ready to be indexed (probably a default
            # value to index from field.emptyIndexValue).
            return value
        elif not hasattr(value, '__iter__'):
            # An object. Get its IID.
            return value.iid
        else:
            # Convert a list of objects into a list of their IIDs
            return [o.iid for o in value]

    @classmethod
    def toTerm(class_, value, field):
        '''Ensure p_value is an object's IID'''
        return value if isinstance(value, int) else value.iid

    @classmethod
    def toString(class_, o, value):
        '''The string representation for tied objects stored in p_value is the
           concatenation of their titles.'''
        return ' '.join([tied.getShownValue() for tied in value]) if value \
                                                                  else None
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from http import HTTPStatus

from appy.px import Px
from appy.ui.js import Quote
from appy.utils import Traceback
from appy.ui.template import Template
from appy.utils import MessageException

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Error:
    '''Represents a server error'''

    byCode = {
     # Error 404: page not found
     404: {'label': 'not_found', 'px':
       Px('''<div>::msg</div>
             <div if="not isAnon and not popup">
              <a href=":tool.computeHomePage()">:_('app_home')</a>
             </div>''', template=Template.px, hook='content')},

     # Error 403: unauthorized
     403: {'label': 'unauthorized', 'px':
       Px('''<div><img src=":url('password.svg')" class="iconERR"/>
                  <x>::msg</x></div>''', template=Template.px, hook='content')},

     # Error 500 server error
     500: {'label': 'server_error', 'px':
       Px('''<div><img src=":url('warning.svg')" class="iconERR"/>
                  <x>::msg</x></div>''', template=Template.px, hook='content')},

     # Error 503: service unavailable. This error code is used by Appy when, due
     # to too many conflict errors (heavy load), the server cannot serve the
     # request.
     503: {'label': 'conflict_error',
           'px': Px('''<div>::msg</div>''',
                    template=Template.px, hook='content')},

     # Return code 200: a logical error rendering a nice translated message to
     # the user, but not considered being an error at the HTTP level.
     200: {'px': Px('''<img src=":url('warning.svg')" class="iconERR"/>
                       <x>::msg</x>''', template=Template.px, hook='content')}
    }

    @classmethod
    def getContent(class_, traversal, code, error):
        '''Get the textual content to render in the error page'''
        # 
        if traversal.user.hasRole('Manager') and (code != 200):
            r = Traceback.get(html=True)
        else:
            if isinstance(error, MessageException):
                r = str(error)
            else:
                _ = traversal.handler.tool.translate
                r = _(Error.byCode[code]['label'])
                if error:
                    text = str(error).strip()
                    if text:
                        r += '<div class="discreet">%s</div>' % text
        return r

    @classmethod
    def get(class_, resp, traversal, error=None):
        '''A server error just occurred. Try to return a nice error page. If it
           fails (ie, the error is produced in the main PX template), dump a
           simple traceback.'''
        # When managing an error, ensure no database commit will occur
        handler = traversal.handler
        handler.commit = False
        # Log the error
        code = resp.code if resp.code in Error.byCode else 500
        # Message exceptions are special errors being 200 at the HTTP level
        is200 = code == 200
        message = '%d on %s' % (code, handler.path)
        if error and error.args:
            message = '%s - %s' % (message, str(error))
        handler.log('app', 'warning' if is200 else 'error', message=message)
        if code in (500, 503):
            handler.log('app', 'error', Traceback.get().strip())
        # Compute the textual content that will be shown
        content = Error.getContent(traversal, code, error)
        # If we are called by an Ajax request, return only the error message,
        # and set the return code to 200; else, browsers will complain.
        context = traversal.context
        if context and context.ajax:
            resp.code = HTTPStatus.OK
            return '<p>%s</p>' % content
        # Return the PX corresponding to the error code. For rendering it, get
        # the last PX context, or create a fresh one if there is no context.
        if not context:
            # Remove some variables from the request to minimise the possibility
            # that an additional error occurs while creating the PX context.
            req = handler.req
            if 'search' in req: del(req.search)
            traversal.context = context = traversal.createContext()
        else:
            # Reinitialise PX counts. Indeed, we have interrupted a "normal"
            # page rendering that has probably already executed a series of PXs.
            # Without reinitialising counts, Appy will believe these PXs were
            # already executed and will not include CSS and JS code related to
            # these PXs.
            context['_rt_'] = {}
        context.msg = content
        try:
            return Error.byCode[code]['px'](context)
        except Exception as err:
            return '<p>%s</p>' % content
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy.fields import Field
from appy.px import Px
from appy.gen.layout import Table

# ------------------------------------------------------------------------------
class Boolean(Field):
    '''Field for storing boolean values'''
    yesNo = {'true': 'yes', 'false': 'no', True: 'yes', False: 'no'}
    trueFalse = {True: 'true', False: 'false'}

    # Default layout (render = "checkbox") ("b" stands for "base"), followed by
    # the 'grid' variant (if the field is in a group with 'grid' style).
    bLayouts = {'view': 'lf', 'edit': Table('f;lrv;-', width=None),
                'search': 'l-f'}
    gLayouts = {'view': 'fl', 'edit': Table('frvl', width=None),
                'search': 'l-f'}
    # Layout including a description (followed by grid variant)
    dLayouts = {'view': 'lf', 'edit': Table('flrv;=d', width=None)}
    # *d*escription also visible on "view"
    dvLayouts = {'view': 'lf-d', 'edit': dLayouts['edit']}
    gdLayouts = {'view': 'lf', 'edit': Table('f;dv-', width=None)}
    # Centered layout, no description
    cLayouts = {'view': 'lf|', 'edit': 'flrv|'}
    # Layout for radio buttons (render = "radios")
    rLayouts = {'edit': 'f', 'view': 'f', 'search': 'l-f'}
    rlLayouts = {'edit': 'l-f', 'view': 'lf', 'search': 'l-f'}
    grlLayouts = {'edit': 'fl', 'view': 'fl', 'search': 'l-f'}

    pxView = pxCell = Px('''
    <x>::field.getInlineEditableValue(obj, value, layoutType, name=name)</x>
    <input type="hidden" if="masterCss"
           class=":masterCss" value=":rawValue" name=":name" id=":name"/>''')

    pxEdit = Px('''<x var="isTrue=field.isTrue(zobj, name, rawValue);
                           visibleName=name + '_visible'">
     <x if="field.render == 'checkbox'">
      <input type="checkbox" name=":visibleName" id=":name"
             class=":masterCss" checked=":isTrue"
             onclick=":field.getOnChange(zobj, layoutType)"/>
     </x>
     <x if="field.render == 'radios'"
        var2="falseId='%s_false' % name;
              trueId='%s_true' % name">
      <input type="radio" name=":visibleName" id=":falseId" class=":masterCss"
             value="False" checked=":not isTrue"
             onclick=":field.getOnChange(zobj, layoutType)"/>
      <label lfor=":falseId">:_(field.labelId + '_false')</label><br/>
      <input type="radio" name=":visibleName" id=":trueId" class=":masterCss"
             value="True" checked=":isTrue"
             onclick=":field.getOnChange(zobj, layoutType)"/>
      <label lfor=":trueId">:_(field.labelId + '_true')</label>
     </x>
     <input type="hidden" name=":name" id=":'%s_hidden' % name"
            value=":isTrue and 'True' or 'False'"/>

     <script if="hostLayout" var2="x=zobj.setLock(user, field=field)">:\
       'prepareForAjaxSave(%s,%s,%s,%s)' % \
        (q(name),q(obj.id),q(obj.url),q(hostLayout))</script></x>''',

     js='''
      updateHiddenBool = function(elem) {
        // Determine the value of the boolean field
        var value = elem.checked,
            hiddenName = elem.name.replace('_visible', '_hidden');
        if ((elem.type == 'radio') && endsWith(elem.id,'_false')) value= !value;
        value = (value)? 'True': 'False';
        // Set this value in the hidden field
        document.getElementById(hiddenName).value = value;
      }''')

    pxSearch = Px('''
      <x var="valueId='%s_yes' % name">
       <input type="radio" value="True" name=":widgetName" id=":valueId"/>
       <label lfor=":valueId">:_(field.getValueLabel(True))</label>
      </x>
      <x var="valueId='%s_no' % name">
       <input type="radio" value="False" name=":widgetName" id=":valueId"/>
       <label lfor=":valueId">:_(field.getValueLabel(False))</label>
      </x>
      <x var="valueId='%s_whatever' % name">
       <input type="radio" value="" name=":widgetName" id=":valueId"
              checked="checked"/>
       <label lfor=":valueId">:_('whatever')</label>
      </x><br/>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts = None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      specificReadPermission=False, specificWritePermission=False, width=None,
      height=None, maxChars=None, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, sdefault=False, scolspan=1, swidth=None, sheight=None,
      persist=True, render='checkbox', inlineEdit=False, view=None, cell=None,
      edit=None, xml=None, translations=None):
        # By default, a boolean is edited via a checkbox. It can also be edited
        # via 2 radio buttons (p_render="radios").
        self.render = render
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          searchable, specificReadPermission, specificWritePermission, width,
          height, None, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        self.pythonType = bool

    def getDefaultLayouts(self):
        if self.render == 'radios': return Boolean.rLayouts
        return self.inGrid() and Boolean.gLayouts or Boolean.bLayouts

    def getValue(self, obj, name=None, layout=None):
        '''Never returns "None". Returns always "True" or "False", even if
           "None" is stored in the DB.'''
        value = Field.getValue(self, obj, name, layout)
        if value is None: return False
        return value

    def getValueLabel(self, value):
        '''Returns the label for p_value (True or False): if self.render is
           "checkbox", the label is simply the translated version of "yes" or
           "no"; if self.render is "radios", there are specific labels.'''
        if self.render == 'radios':
            return '%s_%s' % (self.labelId, self.trueFalse[value])
        return self.yesNo[value]

    def getFormattedValue(self, obj, value, layoutType='view',
                          showChanges=False, language=None):
        return obj.translate(self.getValueLabel(value), language=language)

    def getMasterTag(self, layoutType):
        '''The tag driving slaves is the hidden field'''
        return (layoutType == 'edit') and ('%s_hidden' % self.name) or self.name

    def getOnChange(self, zobj, layoutType, className=None):
        '''Updates the hidden field storing the actual UI value for the field'''
        r = 'updateHiddenBool(this)'
        # Call the base behaviour
        base = Field.getOnChange(self, zobj, layoutType, className=className)
        return base and '%s; %s' % (r, base) or r

    def getStorableValue(self, obj, value, complete=False):
        if not self.isEmptyValue(obj, value):
            exec 'r = %s' % value
            return r

    def getSearchValue(self, form):
        '''Converts the raw search value from p_form into a boolean value'''
        r = Field.getSearchValue(self, form)
        exec 'r = %s' % r
        return r

    def isSortable(self, usage):
        '''Can this field be sortable ?'''
        if usage == 'search': return Field.isSortable(self, usage)
        return True # Sortable in Ref fields

    def isTrue(self, obj, name, dbValue):
        '''When rendering this field as a checkbox, must it be checked or
           not?'''
        rq = obj.REQUEST
        # Get the value we must compare (from request or from database)
        if rq.has_key(name):
            return rq.get(name) in ('True', 1, '1')
        return dbValue
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy import Object
from dav import Resource

# Error management -------------------------------------------------------------
class GoogleError(Exception): pass
SERVER_ERROR = 'Server error: %s.'
STATUS_ERROR = 'Status error: %s.'
RESOURCE_ERROR = 'Resource error: %s.'

# ------------------------------------------------------------------------------
class Address:
    '''Represents an address as geolocalized by Google'''
    def __init__(self, result):
        self.address = result.formatted_address
        coordinates = result.geometry.location
        self.coordinates = coordinates.lat, coordinates.lng

    def __repr__(self):
        return '%s (%f,%f)' % (self.address,
                               self.coordinates[0], self.coordinates[1])

# ------------------------------------------------------------------------------
class Google:
    '''Represents the Google Web Services API'''
    Error = GoogleError

    # URLs to the currently supported Google APIs
    mapsApi = 'https://maps.googleapis.com/maps/api'
    apis = Object(
      geocoding='%s/geocode' % mapsApi,
      distance='%s/distancematrix' % mapsApi
    )

    @staticmethod
    def _call(logger, url):
        '''Sends a HTTP request to Google'''
        logger.log('ask Google: %s' % url)
        server = Resource(url)
        try:
            response = server.get()
        except Resource.Error, re:
            msg = RESOURCE_ERROR % str(re)
            logger.log(msg, type='error')
            raise GoogleError(msg)
        # We got a response
        if response.code != 200:
            msg = SERVER_ERROR % response.text
            logger.log(msg, type='error')
            raise GoogleError(msg)
        # Check the return status
        status = response.data['status']
        if status != 'OK':
            msg = STATUS_ERROR % status
            logger.log(msg, type='error')
            raise GoogleError(msg)
        return response.data

    @staticmethod
    def geocode(key, logger, address, verbose=False):
        '''Returns the coordinates of some given p_address. If p_verbose is
           False, it returns a tuple ~(f_latitude, f_longitude)~. Else, it
           returns the complete data structure as described in the Google
           API.'''
        # Encode p_address
        address = address.replace(' ', '+').replace('\t', '')
        # Perform the HTTP request
        url = '%s/json?address=%s&key=%s' % (Google.apis.geocoding, address,key)
        data = Google._call(logger, url)
        # Return the complete response when required
        if verbose: return data
        # Return the coordinates and formatted addresses only
        results = data.results
        if len(results) == 1: # A single match
            res = Address(results[0])
            msg = 'got one match: %s' % res
        else:
            res = [Address(result) for result in results]
            msg = 'got %d matches.' % len(res)
        logger.log(msg)
        return res

    @staticmethod
    def distance(key, logger, origin, destination, mode='driving',
                 language='fr', verbose=False):
        '''Returns the distance between p_origin and p_destination (expressed
           as tuples of floats (latitude, longitude).'''
        # Encode parameters
        params = 'origins=%s&destinations=%s&mode=%s&language=%s&key=%s' % \
                 (str(origin)[1:-1].replace(' ', ''),
                  str(destination)[1:-1].replace(' ', ''), mode, language, key)
        # Perform the HTTP request
        url = '%s/json?%s' % (Google.apis.distance, params)
        data = Google._call(logger, url)
        # Return the complete response when required
        if verbose: return data
        res = data.rows[0].elements[0].distance.value
        logger.log('response is %dm' % res)
        return res
# ------------------------------------------------------------------------------



'''A "request" stores HTTP GET request parameters and/or HTTP POST form
   values.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from io import BytesIO
from http import HTTPStatus
import os.path, time, email.utils, urllib.parse, gzip

from appy.xml import xhtmlPrologue

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
BROKEN_PIPE = 'Broken pipe while serving %s.'
CONN_RESET  = 'Connection reset by peer while serving %s.'
OS_ERROR    = 'OS error (%s) while serving %s.'
RESP_ERR    = '%d - %s'
H_METH_KO   = 'Unsupported HTTP method "%s".'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Response:
    '''Represents the response that will be sent back to the client after having
       received a HTTP request.'''

    # Appy works with UTF-8
    encoding = 'utf-8'
    charset = 'charset=%s' % encoding

    # HTTP headers must be ISO-8859-1
    headersEncoding = 'iso-8859-1'

    # End of lines as written in HTTP responses
    eol = b'\r\n'

    # Types of dynamic responses a Appy server may produce
    types = {'html': 'text/html;%s' % charset,
             'xml' : 'text/xml;%s' % charset,
             'json': 'application/json;%s' % charset}

    # MIME type of files stored on disk, that will be compressed. CSS files are
    # not there, because the Appy server loads it in RAM.
    compressableFiles = {'image/svg+xml': None, 'application/javascript': None}

    # HTTP codes for which responses have no body
    noBodyCodes = (HTTPStatus.NO_CONTENT,
                   HTTPStatus.RESET_CONTENT,
                   HTTPStatus.NOT_MODIFIED)

    # Template error page
    errorTemplate = '''%s
     <html>
      <head><meta http-equiv="Content-Type" content="%s"><title>Error</title>
      </head>
      <body><p>Error code %%d</p></body>
     </html>''' % (xhtmlPrologue, types['html'])

    # When writing large amounts of data on the client socket, packets of BYTES
    # bytes will be written.
    BYTES = 60000

    def __init__(self, handler):
        # A link to the p_handler
        self.handler = handler
        # The HTTP code for the response, as a value of enum http.HTTPStatus
        self.code = HTTPStatus.OK
        # The phrase related to the code. Ie, phrase for code 200 is "OK".
        self.codePhrase = None
        # The message to be returned to the user
        self.message = None
        # Response base headers, returned for any response, be it static or
        # dynamic.
        if handler.fake:
            headers = {}
        else:
            headers = {
              'Server': handler.server.nameForClients,
              'Date': email.utils.formatdate(time.time(), usegmt=True),
              # For now, disable byte serving (value "bytes" instead of "none")
              'Accept-Ranges': 'none'}
        self.headers = headers
        # If the response content type is among Response.types, the
        # corresponding key in this dict is stored here.
        self.contentType = None
        # The HTTP version string
        self.httpVersion = handler.server.config.server.getProtocolString()
        # Has this response already been sent ? It may happen when producing
        # dynamic content whose result is a File.
        self.sent = False

    def setContentType(self, type):
        '''Sets the content type for this response'''
        self.headers['Content-Type'] = Response.types.get(type) or type
        self.contentType = type

    def initDynamic(self, contentType='html'):
        '''Complete p_self's headers with those being specific to a dynamic
           response.'''
        headers = self.headers
        self.setContentType(contentType)
        headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
        headers['Expires'] = '0'
        # If it turns out that the dynamic content must be cached (it could
        # happen with File fields with cache=True), keys "Cache-Control" and
        # "Expires" will be removed afterwards.

    def setHeader(self, name, value):
        '''Adds (or replace) a HTTP header among response headers'''
        self.headers[name] = value

    def removeHeader(self, name):
        '''Ensure the header named p_name is absent from response headers'''
        if name in self.headers:
            del(self.headers[name])

    def setCookie(self, name, value):
        '''Adds a cookie among response headers, in special key "Cookies" that
           will be converted to as many "Set-Cookie" HTTP header entries as
           there are entries at this key.

           If p_value is "deleted", the cookie is built in such a way that it
           will be immediately disabled by the browser.'''
        # Create entry "Cookies" in response headers if it does not exist
        if 'Cookies' not in self.headers:
            self.headers['Cookies'] = {}
        # Set the value for the cookie. A special value is defined if the
        # objective is to disable the cookie.
        if value == 'deleted': value = '%s; Max-Age=0' % value
        self.headers['Cookies'][name] = '%s; Path=/' % value

    def deleteCookie(self, name):
        '''Deletes the cookie having this p_name'''
        self.setCookie(name, 'deleted')

    def addMessage(self, message):
        '''Adds a message to p_self.message'''
        if self.message is None:
            self.message = message
        else:
            self.message = '%s<br/>%s' % (self.message, message)

    def goto(self, url=None, message=None):
        '''Redirect the user to p_url'''
        if message: self.addMessage(message)
        self.code = HTTPStatus.SEE_OTHER # 303
        # Redirect to p_url or to the referer URL if no p_url has been given
        self.headers['Location'] = url or self.handler.headers['Referer']

    def write(self, s):
        '''Writes part p_s of the response to the client socket. Returns True if
           an error occurred.'''
        try:
            self.handler.clientSocket.sendall(s)
        except BrokenPipeError:
            self.handler.log('app', 'error', BROKEN_PIPE % self.handler.path)
            return True
        except ConnectionResetError:
            self.handler.log('app', 'error', CONN_RESET % self.handler.path)
            return True
        except OSError as err:
            self.handler.log('app', 'error', OS_ERROR % \
                                             (str(err), self.handler.path))

    def writeFile(self, path):
        '''Writes the file lying on the file system @ path to the client socket.
           Return True if an error occurred.'''
        r = False
        with open(path, 'rb') as f:
            while True:
                chunk = f.read(Response.BYTES)
                if not chunk: break
                error = self.write(chunk)
                if error:
                    r = True
                    break
        return r

    def formatHeaderLine(self, line):
        '''Return p_line as can be written as a response header'''
        r = line.encode(Response.headersEncoding, 'ignore')
        return b'%s%s' % (r, Response.eol)

    def writeHeaders(self):
        '''Writes, to the client socket, the response line and headers'''
        # Define the response line
        phrase = self.codePhrase or self.code.phrase or ''
        line = self.formatHeaderLine
        r = [line('%s %d %s' % (self.httpVersion, self.code, phrase))]
        add = r.append
        # Add HTTP headers
        for name, value in self.headers.items():
            # Manage special key containing cookies
            if name == 'Cookies':
                for key, v in value.items():
                    add(line('Set-Cookie: %s=%s' % (key, v)))
            else:
                # Manage any other key
                add(line('%s: %s' % (name, value)))
        # Terminate with a blank line
        add(Response.eol)
        # Write the complete result to the socket
        self.write(b''.join(r))

    def compress(self, content):
        '''Compresses p_content and add the appropriate response header
           "Content-Encoding". The compressed version of p_content is
           returned.'''
        # It must be compressed
        r = BytesIO()
        with gzip.GzipFile(fileobj=r, mode='wb') as f: f.write(content)
        self.headers['Content-Encoding'] = 'gzip'
        return r.getvalue()

    def prepareContent(self, content, path):
        '''Prepares the response body, that will be made of p_content (being a
           string) or the content of file at this p_path.'''
        # Returns a tuple (p_content, p_path).
        #  * If the file @p_path must be compressed, it will be loaded in
        #    p_content. Else, it will be left as is, and will be written to the
        #    client socket afterwards, by chunks of Response.BYTES bytes.
        #  * p_content will be compressed if being "textual".
        # This method also sets header key "Content-Length" accordingly.
        headers = self.headers
        toCompress = False
        type = headers.get('Content-Type')
        # Convert p_content to bytes if present
        if content:
            content = content.encode(Response.encoding)
        elif path and (type in Response.compressableFiles):
            # Loads the file @ p_path in p_content if it must be compressed
            with open(path, 'rb') as f: content = f.read()
            path = None
            toCompress = True
        # Determine header Content-Length in all cases
        if path:
            length = os.path.getsize(path)
        elif not content:
            # There is no content at all
            length = 0
        else:
            # Compress content when relevant and set header "Content-Length"
            # Compress p_content when appropriate
            if toCompress or \
               (type and (type.startswith('app') or type.startswith('text'))):
                content = self.compress(content)
            length = len(content)
        headers['Content-Length'] = length
        return content, path

    def build(self, content=None, path=None):
        '''Build and sent the response back to the browser'''
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # (a) | p_content is not None:
        #     |
        #     | p_content will be used as response body: it is supposed to be a
        #     | UTF-8 string and will be converted to bytes.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # (b) | p_path is not None:
        #     |
        #     | the content of the file at this (absolute) path will be used as
        #     | response body.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # (c) | p_content is None and p_path is None:
        #     |
        #     | the response will contain no body.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # This response may already have been sent. Do not sent it once more.
        if self.sent: return
        # Add p_self.message as a cookie if present
        handler = self.handler
        if self.message is not None:
            quoted = urllib.parse.quote(self.message)
            self.setCookie('AppyMessage', quoted)
        # Prepare the response content
        content, path = self.prepareContent(content, path)
        # Add key "Connection" when appropriate
        if handler.closeSocket:
            self.headers['Connection'] = 'close'
        # Send response headers
        self.writeHeaders()
        # Send response content when present. Variable "error" will be true if
        # an error occurs while writing the response body to the client socket.
        error = False
        if content: error = self.write(content)
        elif path:  error = self.writeFile(path)
        # Ensure the client socket will be closed if an error occurred
        if error:
            handler.closeSocket = True
        # Note this response as already been sent
        self.sent = True

    def buildError(self, code, message=None):
        '''Updates p_self to represent a HTTP error with this code, build it and
           log it.'''
        self.code = code
        message = message or code.phrase
        handler = self.handler
        # Log the error
        handler.log('app', 'error', RESP_ERR % (code, message))
        # Define a content when relevant
        if (code > 200) and (code not in Response.noBodyCodes):
            content = Response.errorTemplate % code
        else:
            content = None
        # Force the client socket to be closed
        handler.closeSocket = True
        # Build the response
        self.build(content=content)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os
from appy.utils import Version

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Monitoring:
    '''Implements stuff allowing to perform monitoring on a Appy application.
       * URL <yourapp>/config/check can be called to get monitoring info;
       * Configure monitoring parameters by updating attributes of the
         Monitoring instance defined in the Config class.'''

    def __init__(self, forceComplete=False):
        # When returning a success status code, what code to return ?
        self.ok = 'OK'
        # When returning a failure status code, what code to return ?
        self.ko = 'KO'
        # Do we check the presence of LibreOffice running in server mode ?
        self.checkLo = True
        # Normally, producing a complete or summary status depends on a request
        # parameter. If p_forceComplete is True, however, we always return the
        # complete status.
        self.forceComplete = forceComplete
        # The app may define here additional version information, as a list of
        # (key, value) tuples, keys and values being strings.
        self.app = []

    def asText(self, r):
        '''Returns monitoring info p_r as pure text'''
        return '\n'.join('%s: %s' % (k, v) for k, v in r)

    def asHtml(self, r):
        '''Returns monitoring info p_r as HTML'''
        rows = ['<tr><th>%s</th><td>%s</td>' % (k, v) for k, v in r]
        return '<table class="grid">%s</table>' % '\n'.join(rows)

    def get(self, request, config, html=False):
        '''Returns monitoring-related info'''
        # The global monitoring status
        success = True
        # Check if LibreOffice is running
        if self.checkLo:
            loLine = ''
            null, out = os.popen4('ps -ef | grep "soffice"')
            for line in out.readlines():
                if "accept=socket" in line:
                    loLine = line
                    break
            if not loLine:
                success = False
        # Do we need to return complete information or only a status code ?
        status = success and self.ok or self.ko
        if ('all' not in request) and not self.forceComplete: return status
        # Return complete information
        r = [('Status', status), ('Appy version', appy.getVersion())]
        if self.checkLo:
            # Appy parameters for connecting to LibreOffice in server mode
            r.append( ('UNO-enabled Python',
                       config.unoEnabledPython or '<not specified>'))
            r.append(('LibreOffice port', str(config.libreOfficePort)))
            # Info about the running LibreOffice server
            if not loLine:
                r.append(('LibreOffice status', 'Not found'))
            else:
                r.append(('LibreOffice status', 'Running: %s' % loLine))
        # Add app-specific info when available
        if self.app: r += self.app
        # Return pure text of HTML when requested
        return html and self.asHtml(r) or self.asText(r)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
UPLOADED = 'file "%s" uploaded in Ref "%d.%s" via rich "%s".'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import sys
from xml.sax._exceptions import SAXParseException

from appy.px import Px
from appy.ui.layout import Layouts
from appy.model.fields import Field
from appy.xml.cleaner import Cleaner
from appy.utils.diff import HtmlDiff
from appy.utils.inject import Injector
from appy.model.fields.text import Text
from appy.utils import string as sutils
from appy.ui.layout import Layouts, Layout
from appy.model.fields.multilingual import Multilingual

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
XML_ERROR   = 'Error while reading content of field %s on %s. %s.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Rich(Multilingual, Field):
    '''Field allowing to encode a "rich" text, based on XHTML and external
       editor "ckeditor".'''

    # Some elements are traversable
    traverse = {}

    # Required Javascript files
    cdnUrl = 'https://cdn.ckeditor.com/%s/%s/ckeditor.js'

    # Use this constant to say that there is no maximum size for a string field
    NO_MAX = sys.maxsize

    class Layouts(Layouts):
        '''Rich-specific layouts'''
        baseEdit = Layout('lrv-d-f', css_class='bottomSpaceS')
        b = Layouts(edit=baseEdit, view='l-f')
        c = Layouts(edit=baseEdit, view='lc-f') # Idem but with history
        f = Layouts(edit=baseEdit, view='f') # Field only (no label) on view
        g = Layouts(edit=Layout('d2-f;rv=', width='99%'),
                    view=Layout('fl', width='99%'))
        gc = Layouts(edit=Layout('drv-f', width='99%'),
                     view=Layout('cl-f', width='99%')) # Idem but with history

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this Rich p_field'''
            # Is this field in a grid-style group ?
            inGrid = field.inGrid()
            if field.historized:
                # self.historized can be a method or a boolean. If it is a
                # method, it means that under some condition, historization will
                # be enabled. So we come here also in this case.
                r = 'gc' if inGrid else 'c'
            else:
                r = 'g' if inGrid else 'b'
            return getattr(class_, r)

    # Default ways to render multilingual fields
    defaultLanguagesLayouts = {
      'edit': 'horizontal', 'view': 'horizontal', 'cell': 'vertical'}

    # Override this dict, defined at the Multilingual level
    lgTop = {'view': 1, 'edit': 1}

    # Default styles to use
    defaultStyles = ('p', 'h1', 'h2', 'h3', 'h4')

    # Default custom styles
    customStyles = [
      # When this style is applied, at the XHTML level, it has no effet, but
      # when dumped in a POD, a page break is inserted just after the paragraph
      # having this style.
      {'id': 'div', 'element': 'div',
       'attributes': {'style': 'page-break-after:always'}},
      # Style for a block of text rendered in a monospace font
      {'id': 'pre', 'element': 'pre'}
    ]

    # The name of the index class storing values of this field in the catalog
    indexType = 'RichIndex'

    # Unilingual view and cell
    viewUni = cellUni = Px('''
     <x var="inlineEdit=field.getAttribute(o, 'inlineEdit');
             css=field.getAttribute(o, 'viewCss');
             mayAjaxEdit=inlineEdit and (layout != 'cell') and not \
                       showChanges and guard.mayEdit(o, field.writePermission);
             value=field.injectContent(o, value) if field.inject else value">
      <div if="not mayAjaxEdit"
           class=":css">::value or field.valueIfEmpty</div>
      <x if="mayAjaxEdit" var2="name=lg and ('%s_%s' % (name, lg)) or name">
       <div class=":css" contenteditable="true"
            id=":'%d_%s_ck' % (o.iid, name)">::value or '-'</div>
       <script if="mayAjaxEdit">::field.getJsInlineInit(o, name, lg)</script>
      </x>
     </x>''')

    # Unilingual edit
    editUni = Px('''
     <textarea var="inputId=not lg and name or '%s_%s' % (name, lg)"
       id=":inputId" name=":inputId" cols=":field.getTextareaCols()"
       style=":field.getTextareaStyle()"
       rows=":field.height">:field.getInputValue(inRequest, requestValue, value)
     </textarea>
     <script>::field.getJsInit(o, lg)</script>''')

    search = Px('''
     <input type="text" maxlength=":field.maxChars" size=":field.swidth"
            value=":field.sdefault" name=":widgetName"/>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      filterField=None, readPermission='read', writePermission='write',
      width=None, height=None, maxChars=None, colspan=1, master=None,
      masterValue=None, focus=False, historized=False, mapping=None,
      generateLabel=None, label=None, sdefault='', scolspan=1, swidth=None,
      sheight=None, persist=True, styles=defaultStyles, customStyles=None,
      documents=False, spellcheck=False, languages=('en',),
      languagesLayouts=None, inlineEdit=False, toolbar='Standard', view=None,
      cell=None, edit=None, xml=None, translations=None, inject=False,
      valueIfEmpty='', viewCss='xhtml'):
        # The list of styles that the user will be able to select in the styles
        # dropdown (within CKEditor) is defined hereafter.
        self.styles = styles
        # If you want to add custom, non-standard styles in the above-mentioned
        # dropdown, do not touch attribute "styles", but specify such entries in
        # the following attribute "customStyles". It must be a list or dict of
        # entries; every entry represents a custom style and must be a dict
        # having the following keys and values. Every key and value must be a
        # string.
        # ----------------------------------------------------------------------
        #    "id"    | An identifier for your style, that must be different from
        #            | any standard CKEditor style like h1, h2, p, div, etc;
        # ----------------------------------------------------------------------
        #   "name"   | A translated name for your style, that will appear in the
        #            | dropdown. Do not use any special (ie, accentuated) char
        #            | in the value: prefer the use of an HTML entity for
        #            | defining such char.
        # ----------------------------------------------------------------------
        #  "element" | The HTML tag that will surround the text onto which the
        #            | style will be applied. Do not use "blockquote", it does
        #            | not work. Non-standard tag "address" works well, or any
        #            | standard tag like h1, h2, etc. Note that if you use a
        #            | standard tag, it will work, but if you have also
        #            | activated it in attribute "style" hereabove, you won't be
        #            | able to make the difference in the result between both
        #            | styles because they will produce a similar result.
        # ----------------------------------------------------------------------
        self.customStyles = customStyles or Rich.customStyles
        # By mentioning, in attribute "documents", the name of a Ref field
        # pointing to Document instances (see appy/model/document.py), images
        # upload is actived in this rich field. An illustration can be found in
        # standard rich field appy/model/page.py::content, that uses, on the
        # same class, the Ref field named "documents".
        self.documents = documents
        # Do we run the CK spellchecker ?
        self.spellcheck = spellcheck
        # What toolbar is used ? Possible values are "Standard" or "Simple".
        self.toolbar = toolbar
        # If p_inject is True, the XHTML chunk contained in a rich field is
        # allowed to contain "a" tags of the form:
        #
        #      (a)      <a href="<url>">py: <part_specifier></a>
        #      (b)     <a href="<anything>">px: <object>*<px></a>
        #
        # These links will not be rendered as-is: they will represent as many
        # placeholders for injecting other content.
        #
        #                            ------------
        #                            | Case (a) |
        #                            ------------
        #
        # Each URL must point to a pure textual resource containing Python code.
        # When rendering this link, Appy will download the code file and extract
        # from it the part corresponding to <part_specifier>. Possible code
        # specifiers are the following.
        # ----------------------------------------------------------------------
        #  c_<name>    | This will extract the definition of *c*lass named
        #              | <name>, including its subsequent docstring and comment
        #              | lines, until the first line of not-commented code is
        #              | reached, or a comment starting with "# ~" is
        #              | encountered.
        # ----------------------------------------------------------------------
        #  s_<cn>.<mn> | This will extract the *s*ignature of method named <mn>
        #              | in class named <cn>, its subsequent docstring and
        #              | comment lines, until the first not-commented line of
        #              | code is reached, or a comment starting with "# ~" is
        #              | encountered.
        # ----------------------------------------------------------------------
        #  m_<cn>.<mn> | This will extract the complete method named <mn> (as
        #              | defined in class named <cn>), including its body.
        # ----------------------------------------------------------------------
        #
        #                            ------------
        #                            | Case (b) |
        #                            ------------
        # 
        # This case is used to inject the content of a PX. The content of the
        # "a" tag must start with "px:", followed by a specifier made of 2
        # parts, separated by an asterisk (*):
        # ----------------------------------------------------------------------
        # "object"  | must be the name of an object, class or module being
        #           | available in the current PX context ;
        # ----------------------------------------------------------------------
        #   "px"    | is the name of the PX to use. It must correspond to an
        #           | attribute, available on "object", and be an instance of
        #           | class appy.px.Px.
        # ----------------------------------------------------------------------
        #
        # Note that, when using POD to dump content of Rich fields via the xhtml
        # function, special "a" tags as described hereabove (cases (a) and (b))
        # will not be included by default in the result. In order to enable
        # that, use parameter inject=True, ie:
        #
        #     do ...
        #     from xhtml(o.someRichField, inject=True)
        #
        self.inject = inject
        # The value to show on non-edit layouts if the field is empty
        self.valueIfEmpty = valueIfEmpty
        # When rendering rich content on view or cell layout, standard Appy CSS
        # class named "xhtml" is applied. If you want to specify another class,
        # use the following attribute. A method can also be passed.
        self.viewCss = viewCss
        # ~~~
        # Call the base constructors
        Multilingual.__init__(self, languages, languagesLayouts)
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          None, searchable, filterField, readPermission, writePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        # No max chars by default
        if maxChars is None:
            self.maxChars = Rich.NO_MAX

    def getDiffValue(self, o, value, i=None, language=None):
        '''If p_i is None, p_value is p_self's currently stored value on p_o (or
           its p_language part if p_language is specified) and the method will
           return a version of p_value that includes the cumulative diffs
           between successive versions, from the very first stored version in
           p_o's history to the currently stored p_value.

           If p_i is not None, p_value is p_self's value as stored in the entry
           whose index is p_i+1 in p_o's history, and the method will return a
           diff between it and p_value's subsequent version, either to be found
           in a more recent entry in p_o's history or, failing that, the
           currently stored value on p_o.'''
        r = last = None
        for event in o.history.iter(eventType='Change',chronological=True,i=i):
            # Ignore irrelevant events
            if not event.hasField(self.name, language=language): continue
            # We have a matching event
            if i is not None:
                # Return the diff between p_value and the newer version stored
                # in this event.
                newer = event.getValue(self.name, language)
                iMsg, dMsg = o.history[i+1].getDiffTexts(o)
                return HtmlDiff(value, newer, iMsg, dMsg).get()
            # If we are here, we must produce a cumulative diff
            if r is None:
                # This is the first encountered version
                r = event.getValue(self.name, language)
            else:
                # We need to produce the difference between current result
                # and this version.
                newer = event.getValue(self.name, language)
                iMsg, dMsg = last.getDiffTexts(o)
                r = HtmlDiff(r, newer, iMsg, dMsg).get()
            # Remember the last walked event
            last = event
        if not last:
            # There is no diff to show
            return value
        # Now we need to compare the result with the current version
        iMsg, dMsg = last.getDiffTexts(o)
        return HtmlDiff(r, value or '', iMsg, dMsg).get()

    def getUniFormattedValue(self, o, value, layout='view', showChanges=False,
                             language=None, contentLanguage=None):
        '''Returns the formatted variant of p_value. If p_contentLanguage is
           specified, p_value is the p_contentLanguage part of a multilingual
           value.'''
        if Field.isEmptyValue(self, o, value) and not showChanges: return ''
        r = value
        if showChanges:
            # Compute the successive changes that occurred on p_value
            r = self.getDiffValue(o, r, language=contentLanguage)
        return r

    def validateUniValue(self, o, value):
        '''Ensure p_value as will be stored (=cleaned) is valid XHTML'''
        try:
            Cleaner().clean(value)
        except SAXParseException as err:
            return o.translate('bad_xml', mapping={'error': str(err)})

    def getSearchValue(self, req, value=None):
        '''The input widget for a Rich field is of type "text". So behave like a
           Text field in that case.'''
        return Text.computeSearchValue(self, req, value=value)

    def getHistoryValue(self, o, value, i, language=None, empty='-'):
        '''For a Rich field, instead of showing the previous p_value in p_o's
           history, show a XHTML diff between p_value and a newer version of it
           in p_o's history before p_i, or, if not found, the value as currently
           stored on p_o.'''
        r = Field.getHistoryValue(self, o, value, i, language, empty=None)
        return empty if not r else self.getDiffValue(o, value, i-1, language)

    def getUniStorableValue(self, o, value):
        '''Gets the p_value as can be stored in the database within p_obj'''
        if not value: return value
        # Clean the value. When image upload is enabled, ckeditor inserts some
        # "style" attrs (ie for image size when images are resized). So in this
        # case we can't remove style-related information.
        try:
            value = Cleaner().clean(value)
        except SAXParseException as err:
            # Errors while parsing p_value can't prevent the user from storing
            # it.
            o.log(XML_ERROR % (self.name, o.id, str(err)), type='warning')
        # Manage maxChars
        max = self.maxChars
        if max and (len(value) > max): value = value[:max]
        return value

    def injectContent(self, o, value):
        '''When p_self.inject is True, this method replaces every "injection
           link" found in p_value with the appropriate content.'''
        if not self.inject: return value
        return Injector.run(value, o)

    def getTextareaStyle(self):
        '''On "edit", get the content of textarea's "style" attribute'''
        # If the width is expressed as a string, the field width must be
        # expressed in attribute "style" (so returned by this method) and not
        # via attribute "cols" (returned by m_getTextareaCols below).
        return isinstance(self.width, str) and ('width:%s' % self.width) or ''

    def getTextareaCols(self):
        '''When this widget must be rendered as an HTML field of type
           "textarea", get the content of its "cols" attribute.'''
        # Use this attribute only if width is expressed as an integer value
        return isinstance(self.width, int) and self.width or ''

    ckLanguages = {'en': 'en_US', 'pt': 'pt_BR', 'da': 'da_DK', 'nl': 'nl_NL',
                   'fi': 'fi_FI', 'fr': 'fr_FR', 'de': 'de_DE', 'el': 'el_GR',
                   'it': 'it_IT', 'nb': 'nb_NO', 'pt': 'pt_PT', 'es': 'es_ES',
                   'sv': 'sv_SE'}

    def getCkLanguage(self, o, language):
        '''Gets the language for CK editor SCAYT. p_language is one of
           self.languages if the field is multilingual, None else. If p_language
           is not supported by CK, we use english.'''
        if not language:
            language = self.getAttribute(o, 'languages')[0]
        if language in self.ckLanguages: return self.ckLanguages[language]
        return 'en_US'

    def getCkParams(self, o, language):
        '''Gets the base params to set on a rich text field'''
        base = '%s/ckeditor' % o.buildUrl()
        ckAttrs = {'customConfig': '%s/config.js' % base,
                   'contentsCss': '%s/contents.css' % base,
                   'stylesSet': '%s/styles.js' % base,
                   'toolbar': self.toolbar, 'format_tags':';'.join(self.styles),
                   'scayt_sLang': self.getCkLanguage(o, language)}
        if self.width: ckAttrs['width'] = self.width
        if self.height: ckAttrs['height'] = self.height
        if self.spellcheck: ckAttrs['scayt_autoStartup'] = True
        # Add custom styles
        if self.customStyles:
            for style in self.customStyles:
                id = style['id']
                ckAttrs['format_%s' % id] = style
                ckAttrs['format_tags'] += ';%s' % id
        if self.documents and not o.isTemp():
            base = o.url
            # Parameter "type=any" is only useful in order to have a parameter.
            # Indeed, when a file is cut & pasted or dragged & dropped to
            # ckeditor, ckeditor calls the URL with suffix "&responseType=json".
            # If there is no existing parameter, the syntax is incorrect
            # (absence of '?' between the URL and its parameters) and the
            # request ends up with a 404 error.
            ckAttrs['filebrowserUploadUrl'] = '%s/%s/upload?type=any' % \
                                              (base, self.name)
            ckAttrs['filebrowserBrowseUrl'] = '%s/pxField?name=%s&popup=True' \
              '&selector=%d,%s,repl' % (base, self.documents, o.iid, self.name)
        if not o.user.hasRole('Manager'):
            ckAttrs['removeButtons'] = 'Source'
        ck = []
        for k, v in ckAttrs.items():
            if isinstance(v, int): sv = str(v)
            elif isinstance(v, bool): sv = str(v).lower()
            elif isinstance(v, dict): sv = str(v)
            else: sv = '"%s"' % v
            ck.append('%s: %s' % (k, sv))
        return ', '.join(ck)

    def getJs(self, layout, r, config):
        '''A rich field depend on CKeditor'''
        if (layout not in ('edit', 'view')) or \
           ((layout == 'view') and not self.inlineEdit): return
        # Compute the URL to ckeditor CDN
        ckUrl = Rich.cdnUrl % (config.ui.ckVersion, config.ui.ckDistribution)
        if ckUrl not in r: r.append(ckUrl)

    def getJsInit(self, o, language):
        '''Gets the Javascript init code for displaying a rich editor for this
           field (rich field only). If the field is multilingual, we must init
           the rich text editor for a given p_language (among self.languages).
           Else, p_languages is None.'''
        name = self.name if not language else '%s_%s' % (self.name, language)
        return 'CKEDITOR.replace("%s", {%s})' % \
               (name, self.getCkParams(o, language))

    def getJsInlineInit(self, o, name, language):
        '''Gets the Javascript init code for enabling inline edition of this
           field (rich text only). If the field is multilingual, the current
           p_language is given and p_name includes it. Else, p_language is
           None.'''
        id = o.iid
        fieldName = language and name.rsplit('_',1)[0] or name
        lg = language or ''
        return "CKEDITOR.disableAutoInline = true;\n" \
               "CKEDITOR.inline('%d_%s_ck', {%s, on: {blur: " \
               "function( event ) { var content = event.editor.getData(); " \
               "doInlineSave('%d','%s','%s','view',true,content,'%s')}}})" % \
               (id, name, self.getCkParams(o,language), id, fieldName, o.url,lg)

    traverse['upload'] = 'perm:write'
    def upload(self, o):
        '''Uploads a document to the Ref mentioned in attribute "documents"'''
        # Get the file from the request
        ofile = o.req.upload
        # Create a Document instance and link it to p_o via the Ref mentioned in
        # self.documents.
        doc = o.create(self.documents, title=ofile.name, file=ofile)
        o.H().commit = True
        o.resp.setContentType('json')
        r = {'uploaded': 1, 'fileName': doc.title,
             'url': '%s/file/download' % doc.url}
        o.log(UPLOADED % (ofile.name, o.iid, self.documents, self.name))
        return sutils.getStringFrom(r, c='"')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
